# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module contains common commands used by kas plugins.
"""

import tempfile
import logging
import shutil
import os
import pprint
import configparser
import json
import base64
from pathlib import Path
from git.config import GitConfigParser
from .libkas import (ssh_cleanup_agent, ssh_setup_agent, ssh_no_host_key_check,
                     get_build_environ, repos_fetch, repos_apply_patches,
                     add_cachedir_tag)
from .context import ManagedEnvironment as ME
from .context import get_context
from .includehandler import IncludeException
from .kasusererror import EnvSetButNotFoundError, ArgsCombinationError
from .repos import SignatureValidator

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2018'

KAS_USER_NAME = 'kas User'
KAS_USER_EMAIL = 'kas@example.com'
HTTPS_PORT_DEFAULT = 443
SSH_PORT_DEFAULT = 22


class Macro:
    """
        Contains commands and provides method to run them.
    """

    def __init__(self, use_common_setup=True):
        if use_common_setup:
            repo_loop = Loop('repo_setup_loop')
            repo_loop.add(SetupReposStep())

            # setup commands are pairs of setup / cleanup commands
            self.setup_commands = [
                (SetupDir(), None)
            ]

            if ('SSH_PRIVATE_KEY' in os.environ
                    or 'SSH_PRIVATE_KEY_FILE' in os.environ):
                if 'SSH_AUTH_SOCK' in os.environ:
                    raise ArgsCombinationError(
                        'Internal SSH agent (e.g. for "SSH_PRIVATE_KEY") can '
                        'only be started if no external one is passed.')
                self.setup_commands.append((SetupSSHAgent(),
                                            CleanupSSHAgent()))

            self.setup_commands += [(x, None) for x in [
                SetupHome(),
                MakeNonInteractive(),
                InitSetupRepos(),
                repo_loop,
                FinishSetupRepos(),
                ReposCheckout(),
                ReposApplyPatches(),
                SetupEnviron(),
                WriteBBConfig(),
            ]]
        else:
            self.setup_commands = []

        self.commands = []

    def add(self, command):
        """
            Appends commands to the command list.
        """
        self.commands.append(command)

    def run(self, ctx, skip=None):
        """
            Runs a command from the command list with respect to the
            configuration.
        """
        def _run_single(command):
            command_name = str(command)
            if command_name in (skip or []):
                return False
            logging.debug('execute %s', command_name)
            command.execute(ctx)
            return True

        cleanup_commands = []
        try:
            for cmd in self.setup_commands:
                if _run_single(cmd[0]) and cmd[1]:
                    cleanup_commands.insert(0, cmd[1])
            for cmd in self.commands:
                _run_single(cmd)
        finally:
            for cmd in cleanup_commands:
                _run_single(cmd)


class Command:
    """
        An abstract class that defines the interface of a command.
    """

    def execute(self, ctx):
        """
            This method executes the command.
        """
        pass


class Loop(Command):
    """
        A class that defines a set of commands as a loop.
    """

    def __init__(self, name):
        self.commands = []
        self.name = name

    def __str__(self):
        return self.name

    def add(self, command):
        """
            Appends a command to the loop.
        """
        self.commands.append(command)

    def execute(self, ctx):
        """
            Executes the loop.
        """
        loop_name = str(self)

        def executor(command):
            command_name = str(command)
            logging.debug('Loop %s: execute %s', loop_name, command_name)
            return command.execute(ctx)

        while all(executor(c) for c in self.commands):
            pass


class SetupHome(Command):
    """
        Sets up the home directory of kas.
    """

    # A list of environment variables that SETUP_HOME uses
    # This should be kept up to date with any code in execute()
    ENV_VARS = [
        'GIT_CREDENTIAL_HELPER',
        'GIT_CREDENTIAL_USEHTTPPATH',
        'GITCONFIG_FILE',
        'AWS_CONFIG_FILE',
        'AWS_ROLE_ARN',
        'AWS_SHARED_CREDENTIALS_FILE',
        'AWS_WEB_IDENTITY_TOKEN_FILE',
        'NETRC_FILE',
        'NPMRC_FILE',
        'REGISTRY_AUTH_FILE',
    ]

    def __init__(self):
        super().__init__()
        self.tmpdirname = None

    def __del__(self):
        if self.tmpdirname:
            shutil.rmtree(self.tmpdirname)

    def __str__(self):
        return 'setup_home'

    @staticmethod
    def _path_from_env(name):
        """
        Returns the path a env var points to (if set). If the path does not
        exist, raises an EnvSetButNotFoundError.
        """
        rawpath = os.environ.get(name, None)
        if not rawpath:
            return None
        path = Path(rawpath)
        if not path.exists():
            raise EnvSetButNotFoundError(name, path)
        return path

    @staticmethod
    def _ssh_config_present():
        """
            Checks if the .ssh/config file exists or any manual ssh config
            option is set.
        """
        ssh_vars = ['SSH_PRIVATE_KEY', 'SSH_PRIVATE_KEY_FILE', 'SSH_AUTH_SOCK']
        if any(e in os.environ for e in ssh_vars):
            return True

        ssh_path = os.path.expanduser('~/.ssh')
        if os.path.exists(os.path.join(ssh_path, 'config')):
            return True
        return False

    def _setup_netrc(self):
        netrc_file = self._path_from_env('NETRC_FILE')
        if netrc_file:
            shutil.copy(netrc_file, self.tmpdirname + "/.netrc")
        if os.environ.get('CI_SERVER_HOST', False) \
                and os.environ.get('CI_JOB_TOKEN', False):
            with open(self.tmpdirname + '/.netrc', 'a') as fds:
                fds.write('machine ' + os.environ['CI_SERVER_HOST'] + '\n'
                          'login gitlab-ci-token\n'
                          'password ' + os.environ['CI_JOB_TOKEN'] + '\n')

    def _setup_npmrc(self):
        npmrc_file = self._path_from_env('NPMRC_FILE')
        if not npmrc_file:
            return
        shutil.copy(npmrc_file, self.tmpdirname + "/.npmrc")

    def _setup_registry_auth(self):
        os.makedirs(self.tmpdirname + "/.docker")
        reg_auth_file = self._path_from_env('REGISTRY_AUTH_FILE')
        if reg_auth_file:
            shutil.copy(reg_auth_file,
                        self.tmpdirname + "/.docker/config.json")
        elif not os.path.exists(self.tmpdirname + '/.docker/config.json'):
            with open(self.tmpdirname + '/.docker/config.json', 'w') as fds:
                fds.write("{}")

        if os.environ.get('CI_REGISTRY', False) \
                and os.environ.get('CI_JOB_TOKEN', False) \
                and os.environ.get('CI_REGISTRY_USER', False):
            with open(self.tmpdirname + '/.docker/config.json', 'r+') as fds:
                data = json.loads(fds.read())
                token = os.environ['CI_JOB_TOKEN']
                base64_token = base64.b64encode(token.encode()).decode()
                auths = data.get('auths', {})
                auths.update(
                    {os.environ['CI_REGISTRY']: {"auth": base64_token}})
                data['auths'] = auths
                fds.seek(0)
                fds.write(json.dumps(data, indent=4))
                fds.truncate()

    def _setup_aws_creds(self):
        aws_dir = self.tmpdirname + "/.aws"
        conf_file = aws_dir + "/config"
        shared_creds_file = aws_dir + "/credentials"
        sso_cache_dir = aws_dir + "/sso/cache"
        os.makedirs(aws_dir)
        aws_conf_file = self._path_from_env('AWS_CONFIG_FILE')
        aws_shared_creds_file = \
            self._path_from_env('AWS_SHARED_CREDENTIALS_FILE')
        if aws_conf_file and aws_shared_creds_file:
            shutil.copy(aws_conf_file, conf_file)
            shutil.copy(aws_shared_creds_file, shared_creds_file)

        # OAuth 2.0 workflow credentials
        aws_web_identity_token_file = \
            self._path_from_env('AWS_WEB_IDENTITY_TOKEN_FILE')
        if aws_web_identity_token_file and os.environ.get('AWS_ROLE_ARN'):
            webid_token_file = aws_dir + '/web_identity_token'
            config = configparser.ConfigParser()
            if os.path.exists(conf_file):
                config.read(conf_file)
            if 'default' not in config:
                config['default'] = {}
            config['default']['role_arn'] = os.environ.get('AWS_ROLE_ARN')
            config['default']['web_identity_token_file'] = webid_token_file
            with open(aws_dir + '/config', 'w') as fds:
                config.write(fds)
            shutil.copy(aws_web_identity_token_file, webid_token_file)

        # SSO workflow
        if aws_conf_file:
            aws_cache_dir_conf = \
                os.path.join(os.path.dirname(aws_conf_file), "sso/cache")
            aws_cache_dir_home = os.path.join(Path.home(), ".aws/sso/cache")

            # In kas-container the directory passed in --aws-dir is not
            # mounted in $HOME. Look for sso/cache in directory containing
            # AWS_CONFIG_FILE first to maintain the same behavior between
            # kas and kas-container.
            aws_cache_dir = None
            if os.path.isdir(aws_cache_dir_conf):
                aws_cache_dir = aws_cache_dir_conf
            elif os.path.isdir(aws_cache_dir_home):
                aws_cache_dir = aws_cache_dir_home

            if aws_cache_dir:
                shutil.copy(aws_conf_file, conf_file)
                shutil.copytree(aws_cache_dir, sso_cache_dir)

    @staticmethod
    def _setup_gitlab_ci_ssh_rewrite(config):
        ci_host = os.environ.get('CI_SERVER_HOST', None)
        ci_port = os.environ.get('CI_SERVER_PORT', HTTPS_PORT_DEFAULT)
        ci_prot = os.environ.get('CI_SERVER_PROTOCOL', 'https')
        # added in GitLab 15.11. Set sensible defaults for older versions.
        ci_ssh_host = os.environ.get('CI_SERVER_SHELL_SSH_HOST', ci_host)
        ci_ssh_port = os.environ.get('CI_SERVER_SHELL_SSH_PORT',
                                     SSH_PORT_DEFAULT)

        for host in [ci_host, f'{ci_host}:{ci_port}']:
            section = f'url "{ci_prot}://{host}/"'
            config.add_value(section, 'insteadOf',
                             f'git@{ci_ssh_host}:')
            config.add_value(section, 'insteadOf',
                             f'git@{ci_ssh_host}:{ci_ssh_port}')
            config.add_value(section, 'insteadOf',
                             f'ssh://git@{ci_ssh_host}/')
            config.add_value(section, 'insteadOf',
                             f'ssh://git@{ci_ssh_host}:{ci_ssh_port}/')

    def _setup_gitconfig(self):
        gitconfig_host = self._path_from_env('GITCONFIG_FILE')
        gitconfig_kas = self.tmpdirname + '/.gitconfig'

        # when running in a externally managed environment,
        # always try to read the gitconfig
        if not gitconfig_host and get_context().managed_env:
            gitconfig_host = Path('~/.gitconfig').expanduser()
            if not gitconfig_host.exists():
                gitconfig_host = None

        if gitconfig_host:
            shutil.copy(gitconfig_host, gitconfig_kas)

        with GitConfigParser(gitconfig_kas, read_only=False) as config:
            if os.environ.get('GIT_CREDENTIAL_HELPER', False):
                config['credential'] = {
                    'helper': os.environ.get('GIT_CREDENTIAL_HELPER')
                }
                if os.environ.get('GIT_CREDENTIAL_USEHTTPPATH', False):
                    config['credential']['useHttpPath'] = \
                        os.environ.get('GIT_CREDENTIAL_USEHTTPPATH')

            if get_context().managed_env == ME.GITLAB_CI and \
                    not gitconfig_host:
                ci_project_dir = self._path_from_env('CI_PROJECT_DIR')
                if ci_project_dir:
                    logging.debug('Adding git safe.directory %s',
                                  ci_project_dir)
                    config.add_value('safe', 'directory', str(ci_project_dir))

                ci_server = os.environ.get('CI_SERVER_HOST', None)
                if ci_server and not self._ssh_config_present():
                    logging.debug('Adding GitLab CI ssh -> https rewrites')
                    self._setup_gitlab_ci_ssh_rewrite(config)
            config.write()

    def execute(self, ctx):
        managed_env = get_context().managed_env
        if managed_env:
            logging.info(f'Running on {managed_env}')
        if not self.tmpdirname:
            self.tmpdirname = tempfile.mkdtemp()
        def_umask = os.umask(0o077)
        self._setup_netrc()
        self._setup_npmrc()
        self._setup_registry_auth()
        self._setup_gitconfig()
        self._setup_aws_creds()
        os.umask(def_umask)

        ctx.environ['HOME'] = self.tmpdirname


class MakeNonInteractive(Command):
    """
        Make execution environment non-interactive
    """

    def __str__(self):
        return 'make_non_interactive'

    def execute(self, ctx):
        ctx.environ['GIT_TERMINAL_PROMPT'] = 'false'


class MakeInteractive(Command):
    """
        Make execution environment interactive
    """

    def __str__(self):
        return 'make_interactive'

    def execute(self, ctx):
        ctx.environ['GIT_TERMINAL_PROMPT'] = 'true'


class SetupDir(Command):
    """
        Creates the build directory.
    """

    def __str__(self):
        return 'setup_dir'

    def execute(self, ctx):
        if not os.path.exists(ctx.build_dir):
            os.mkdir(ctx.build_dir)
        add_cachedir_tag(ctx.build_dir, "kas build directory")


class SetupSSHAgent(Command):
    """
        Sets up the ssh agent configuration.
    """

    def __str__(self):
        return 'setup_ssh_agent'

    def execute(self, ctx):
        ssh_tools = ['ssh', 'ssh-add', 'ssh-agent']
        for tool in ssh_tools:
            if shutil.which(tool) is None:
                raise RuntimeError('SSH setup requested but could '
                                   f'not find "{tool}" in PATH')
        ssh_setup_agent()
        # disabling host key checking is persistent and not limited to kas,
        # hence only do it in externally managed environments
        if ctx.managed_env:
            ssh_no_host_key_check()


class CleanupSSHAgent(Command):
    """
        Removes all the identities and stops the ssh-agent instance.
    """

    def __str__(self):
        return 'cleanup_ssh_agent'

    def execute(self, ctx):
        ssh_cleanup_agent()


class SetupEnviron(Command):
    """
        Sets up the kas environment.
    """

    def __str__(self):
        return 'setup_environ'

    def execute(self, ctx):
        ctx.environ.update(
            get_build_environ(ctx.config.get_build_system(), ctx)
        )


class WriteBBConfig(Command):
    """
        Writes bitbake configuration files into the build directory.
    """

    def __str__(self):
        return 'write_bbconfig'

    def execute(self, ctx):
        def _get_layer_path_under_topdir(ctx, layer):
            """
                Returns a path relative to ${TOPDIR}.
                TOPDIR is a BB variable pointing to the build directory.
                It is not expanded by KAS, hence we avoid
                absolute paths pointing into the build host.
            """
            # we need to walk up, which would require pathlib.Path >= 3.12
            relpath = os.path.relpath(layer.path, ctx.build_dir)
            return '${TOPDIR}/' + relpath

        def _write_bblayers_conf(ctx):
            filename = ctx.build_dir + '/conf/bblayers.conf'
            if not os.path.isdir(os.path.dirname(filename)):
                os.makedirs(os.path.dirname(filename))
            layers_sorted = \
                sorted([layer for repo in ctx.config.get_repos()
                        for layer in repo.layers])
            with open(filename, 'w') as fds:
                fds.write(ctx.config.get_bblayers_conf_header())
                fds.write('BBLAYERS ?= " \\\n    ')
                fds.write(' \\\n    '.join(
                          [_get_layer_path_under_topdir(ctx, layer)
                           for layer in layers_sorted]))
                fds.write('"\n')
                fds.write('BBPATH ?= "${TOPDIR}"\n')
                fds.write('BBFILES ??= ""\n')

        def _write_local_conf(ctx):
            filename = ctx.build_dir + '/conf/local.conf'
            with open(filename, 'w') as fds:
                fds.write(ctx.config.get_local_conf_header())
                fds.write(f'MACHINE ??= "{ctx.config.get_machine()}"\n')
                fds.write(f'DISTRO ??= "{ctx.config.get_distro()}"\n')
                fds.write('BBMULTICONFIG ?= '
                          f'"{ctx.config.get_multiconfig()}"\n')
                if ctx.config.get_build_system() == 'isar-rootless':
                    fds.write('ISAR_ROOTLESS ?= "1"\n')

        _write_bblayers_conf(ctx)
        _write_local_conf(ctx)


class ReposApplyPatches(Command):
    """
        Applies the patches defined in the configuration to the repositories.
    """

    def __str__(self):
        return 'repos_apply_patches'

    def _vcs_operate_as_kas(self, gitconfig):
        # currently only implemented for git
        with GitConfigParser(gitconfig, read_only=False) as config:
            # in case no user is defined, we keep the kas user
            user_orig = {
                'name': config.get('user', 'name',
                                   fallback=KAS_USER_NAME),
                'email': config.get('user', 'email',
                                    fallback=KAS_USER_EMAIL)
            }
            config.set_value('user', 'name', KAS_USER_NAME)
            config.set_value('user', 'email', KAS_USER_EMAIL)
            config.write()
            return user_orig

    def _vcs_restore_user(self, gitconfig, user):
        # currently only implemented for git
        with GitConfigParser(gitconfig, read_only=False) as config:
            config['user'] = user
            config.write()

    def execute(self, ctx):
        if 'HOME' not in ctx.environ:
            raise ArgsCombinationError('Apply patches requires setup_home')

        gitconfig = ctx.environ['HOME'] + '/.gitconfig'
        user = self._vcs_operate_as_kas(gitconfig)

        repos_apply_patches(ctx.config.get_repos())

        self._vcs_restore_user(gitconfig, user)


class InitSetupRepos(Command):
    """
        Prepares setting up repos including the include logic
    """

    def __str__(self):
        return 'init_setup_repos'

    def execute(self, ctx):
        ctx.missing_repo_names = ctx.config.find_missing_repos()
        ctx.missing_repo_names_old = None


class SetupReposStep(Command):
    """
        Single step of the checkout repos loop
    """

    def __str__(self):
        return 'setup_repos_step'

    def execute(self, ctx):
        if not ctx.missing_repo_names:
            return False

        if ctx.missing_repo_names == ctx.missing_repo_names_old:
            raise IncludeException('Could not fetch all repos needed by '
                                   'includes. Missing repos: {}'
                                   .format(', '.join(ctx.missing_repo_names)))

        logging.debug('Missing repos for complete config:\n%s',
                      pprint.pformat(ctx.missing_repo_names))

        ctx.missing_repos = []
        for repo_name in ctx.missing_repo_names:
            if repo_name not in ctx.config.get_repos_config():
                # we don't have this repo yet (e.g. due to transitive incl.)
                continue
            ctx.missing_repos.append((repo_name,
                                      ctx.config.get_repo(repo_name)))

        repos_fetch([v for k, v in ctx.missing_repos])

        # import keys from old config and validate against this state
        SignatureValidator.import_keys(ctx)
        for _, repo in ctx.missing_repos:
            # check signature prior to checkout and config dict update
            # to avoid TOCTOU issues
            SignatureValidator.ensure_valid_if_signed(ctx, repo)
            repo.checkout()

        ctx.config.repo_dict.update(
            {id: repo for id, repo in ctx.missing_repos})

        repo_paths = {r: ctx.config.repo_dict[r].path for r
                      in ctx.config.repo_dict}
        ctx.missing_repo_names_old = ctx.missing_repo_names

        ctx.missing_repo_names = \
            ctx.config.find_missing_repos(repo_paths)

        return ctx.missing_repo_names


class FinishSetupRepos(Command):
    """
        Finalizes the repo setup loop
    """

    def __str__(self):
        return 'finish_setup_repos'

    def execute(self, ctx):
        # now fetch everything with complete config
        repos_fetch(ctx.config.get_repos())

        config_str = pprint.pformat(ctx.config.get_config(), sort_dicts=False)
        logging.debug('Configuration from config file:\n%s', config_str)


class ReposCheckout(Command):
    """
        Ensures that the right revision of each repo is checked out.
    """

    def __str__(self):
        return 'repos_checkout'

    def execute(self, ctx):
        SignatureValidator.import_keys(ctx)
        for repo in ctx.config.get_repos():
            SignatureValidator.ensure_valid_if_signed(ctx, repo)
            repo.checkout()



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2019
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module contains the Repo class.
"""

import re
import os
import linecache
import logging
import shutil
from pathlib import Path
from dataclasses import dataclass
from datetime import datetime
from urllib.parse import urlparse
from tempfile import TemporaryDirectory

from kas.configschema import CONFIGSCHEMA
from kas.keyhandler import GPGKeyHandler, SSHKeyHandler
from .context import get_context
from .libkas import run_cmd_async, run_cmd
from .kasusererror import KasUserError
from functools import cached_property
from git import GitCommandError, Repo as GitPythonRepo

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2018'


class UnsupportedRepoTypeError(KasUserError, NotImplementedError):
    """
    The requested repo type is unsupported / not implemented
    """
    pass


class RepoRefError(KasUserError):
    """
    The requested repo reference is invalid, missing or could not be found
    """
    pass


class PatchFileNotFound(KasUserError, FileNotFoundError):
    """
    The requested patch file was not found
    """
    pass


class PatchMappingError(KasUserError):
    """
    The requested patch can not be related to a repo
    """
    pass


class PatchApplyError(KasUserError):
    """
    The provided patch file could not be applied
    """
    def __init__(self, msg, cmd=None, out=None, err=None):
        if cmd:
            msg += '\nvcs command: ' + ' '.join(cmd)
        if out and out.strip():
            msg += f'\nvcs output:\n{out.strip()}'
        if err and err.strip():
            msg += f'\nvcs error:\n{err.strip()}'
        super().__init__(msg)


class RepoFetchError(KasUserError):
    """
    An error occurred during repo fetching
    """
    def __init__(self, repo, output):
        super().__init__(f'Fetching of repo: "{repo.name}" failed: {output}')


class Repo:
    """
        Represents a repository in the kas configuration.
    """

    def __init__(self, name, url, path, commit, tag, branch, refspec, layers,
                 patches, signers, disable_operations):
        self.name = name
        self.url = url
        self.path = path
        self.commit = commit
        self.tag = tag
        self.branch = branch
        self.refspec = refspec
        self.layers = layers
        self._patches = patches
        self.allowed_signers = signers
        self.operations_disabled = disable_operations

        if not self.url:
            self.resolve_local()

    @property
    def qualified_name(self):
        url = urlparse(self.url)
        return (f'{url.netloc}{url.path}'
                .replace('@', '.')
                .replace(':', '.')
                .replace('/', '.')
                .replace('*', '.'))

    @property
    def effective_url(self):
        mirrors = os.environ.get('KAS_PREMIRRORS', '')
        for mirror in mirrors.split('\n'):
            try:
                expr, subst = mirror.split()
                if re.match(expr, self.url):
                    return re.sub(expr, subst, self.url)
            except ValueError:
                continue
        return self.url

    @cached_property
    def revision(self):
        if self.commit:
            (_, output) = run_cmd(self.get_commit_cmd(),
                                  cwd=self.path, fail=False)
            if output:
                return output.strip()
            return self.commit
        if self.tag:
            (_, output) = run_cmd(self.resolve_tag_cmd(),
                                  cwd=self.path, fail=False)
            if output:
                return output.strip()
            return self.tag
        branch = self.branch or self.refspec
        if branch:
            (_, output) = run_cmd(self.resolve_branch_cmd(),
                                  cwd=self.path, fail=False)
            if output:
                return output.strip()
            return branch
        return None

    @cached_property
    def revision_annotated_tag(self):
        if not self.tag:
            return None
        try:
            cmd = self.resolve_annotated_tag_cmd()
        except NotImplementedError:
            return None
        (retc, output) = run_cmd(cmd, cwd=self.path, fail=False)
        if retc or not output:
            return None
        return output.strip()

    @cached_property
    def dirty(self):
        if not self.url:
            return True
        (_, output) = run_cmd(self.is_dirty_cmd(),
                              cwd=self.path, fail=False)
        return bool(output)

    @cached_property
    def signers_type(self):
        if self.allowed_signers is None:
            return 'gpg'
        signers = get_context().config.get_signers_config()
        try:
            ktypes = [signers[k].get('type', 'gpg')
                      for k in self.allowed_signers]
        except KeyError as e:
            raise KasUserError(f'Repository {self.name}: '
                               f'Allowed signer "{e}" not found in config')
        if len(set(ktypes)) > 1:
            raise KasUserError(f'Repository {self.name}: '
                               'Mixed signer types are not supported')
        return ktypes[0]

    @property
    def keyhandler(self):
        if not self.signed:
            return None
        return get_context().keyhandler[self.signers_type]

    def check_signature(self):
        self.keyhandler.prepare_validation(self)
        (ret, _, err) = run_cmd(self.is_signed_cmd(),
                                cwd=self.path, fail=False, capture_stderr=True)
        logging.debug('Signature verification output (%d):\n%s', ret, err)
        if ret != 0:
            return (False, None)
        return self.keyhandler.validate_allowed_signer(self, err)

    @cached_property
    def signed(self):
        return self.allowed_signers is not None

    @staticmethod
    def get_type():
        """
            Repo type as defined in spdx-spec/v2.3/package-information
        """
        raise NotImplementedError("Repo type not implemented")

    def contains_path(self, path):
        (ret, _) = run_cmd(self.contains_path_cmd(str(path)),
                           cwd=self.path, fail=False)
        return ret == 0

    def __str__(self):
        if self.commit and (self.tag or self.branch):
            refspec = f'{self.commit}({self.tag or self.branch})'
        else:
            refspec = self.commit or self.tag or self.branch or self.refspec

        return f'{self.url}:{refspec} ' \
               f'{self.path} {self.layers}'

    __legacy_refspec_warned__ = []
    __no_commit_warned__ = []

    @staticmethod
    def factory(name, repo_config, repo_defaults, repo_fallback_path,
                repo_overrides={}):
        """
            Returns a Repo instance depending on parameters.
            This factory function is referential transparent.
        """
        default_patch_repo = repo_defaults.get('patches', {}).get('repo', None)
        patches_dict = repo_config.get('patches', {})
        patches = []
        for p in sorted(patches_dict):
            if not patches_dict[p]:
                continue
            this_patch = {
                'id': p,
                'repo': patches_dict[p].get('repo', default_patch_repo),
                'path': patches_dict[p]['path'],
            }
            if this_patch['repo'] is None:
                raise PatchMappingError(
                    f'No repo specified for patch entry "{p}" and no '
                    'default repo specified.')

            patches.append(this_patch)

        url = repo_config.get('url', None)
        name = repo_config.get('name', name)
        repo_type = repo_config.get('type', 'git')
        commit = repo_config.get('commit', None)
        tag = repo_config.get('tag', repo_defaults.get('tag', None))
        branch = repo_config.get('branch', repo_defaults.get('branch', None))
        refspec = repo_config.get('refspec',
                                  repo_defaults.get('refspec', None))
        if commit is None and tag is None and branch is None \
                and refspec is None and url is not None:
            raise RepoRefError('No commit, tag or branch specified for '
                               f'repository "{name}". This is only allowed '
                               'for local repositories.')
        if refspec is None:
            commit = repo_overrides.get('commit', commit)
            branch = repo_overrides.get('branch', branch)
            url = repo_overrides.get('url', url)
            if commit and get_context().update:
                logging.warning(f'Update of "{name}" requested, but repo is '
                                'pinned to a fixed commit. Not updating.')
        else:
            if name not in Repo.__legacy_refspec_warned__:
                logging.warning('Using deprecated refspec for repository "%s".'
                                ' You should migrate to commit/tag/branch.',
                                name)
                Repo.__legacy_refspec_warned__.append(name)
            if commit is not None or tag is not None or branch is not None:
                raise RepoRefError(
                    'Unsupported mixture of legacy refspec and '
                    f'commit/tag/branch for repository "{name}"')
            refspec = repo_overrides.get('commit', refspec)
        if url and not commit and name not in Repo.__no_commit_warned__ \
                and get_context().unpinned_repo_warnings:
            if tag:
                logging.warning('Using tag without commit for repository '
                                '"%s" is unsafe as tags are mutable.', name)
                Repo.__no_commit_warned__.append(name)
            elif branch:
                logging.warning('Using branch without commit for repository '
                                '"%s" is unsafe. Either add a commit or use '
                                'a lock file.', name)
                Repo.__no_commit_warned__.append(name)
        path = repo_config.get('path', None)
        signed = repo_config.get('signed', False)
        signers = repo_config.get('allowed_signers', None) if signed else None
        if signed and not signers:
            raise KasUserError(f'Repository "{name}" is signed but no allowed '
                               'signers specified.')
        disable_operations = False

        if path is None:
            if url is None:
                path = Repo.get_root_path(repo_fallback_path)
                logging.info('Using %s as root for repository %s', path, name)
            else:
                path = os.path.join(get_context().kas_work_dir, name)
        elif not os.path.isabs(path):
            # Relative pathes are assumed to start from work_dir
            path = os.path.join(get_context().kas_work_dir, path)

        layers_dict = repo_config.get('layers', {'': None})
        layers = Repo._create_layers_from_dict(name, path, layers_dict)

        if url is None:
            # No version control operation on repository
            disable_operations = True

        if repo_type == 'git':
            if commit and not re.match(r'^[0-9a-f]{40}|[0-9a-f]{64}$', commit):
                logging.warning(
                    f'{commit} is not a full-length hash for repo '
                    f'"{name}". This will be an error in future versions.')
            return GitRepo(name, url, path, commit, tag, branch, refspec,
                           layers, patches, signers, disable_operations)
        if repo_type == 'hg':
            if not shutil.which('hg'):
                raise UnsupportedRepoTypeError(
                    'hg is required for Mercurial repositories')
            return MercurialRepo(name, url, path, commit, tag, branch, refspec,
                                 layers, patches, signers, disable_operations)
        raise UnsupportedRepoTypeError(f'Repo type "{repo_type}" '
                                       'not supported.')

    @staticmethod
    def get_root_path(path, fallback=True):
        """
            Checks if path is under version control and returns its root path.
            If the repo is a submodule, the root path of the super-repository
            is returned.
        """
        git_cmd = ['git', 'rev-parse', '--show-toplevel',
                   '--show-superproject-working-tree']
        (ret, output) = run_cmd(git_cmd, cwd=path, fail=False)
        if ret == 0:
            return sorted(output.strip().split('\n'))[0]

        (ret, output) = run_cmd(['hg', 'root'],
                                cwd=path, fail=False)
        if ret == 0:
            return output.strip()

        return path if fallback else None

    @staticmethod
    def _create_layers_from_dict(repo_name, repo_path, layers_dict):
        layers = []
        disabled_token = "disabled"
        legacy_disabled_tokens = ['excluded', 'n', 'no', '0', 'false', 0]
        default_prio = \
            CONFIGSCHEMA['$defs']['layerPrio']['properties']['prio']['default']

        for lname, prop in layers_dict.items():
            if prop is None:
                layers.append(RepoLayer(name=lname,
                                        repo_path=Path(repo_path),
                                        repo_name=repo_name,
                                        priority=default_prio))
            elif isinstance(prop, str) and prop == disabled_token:
                continue
            elif isinstance(prop, (str, int)) and \
                    prop in legacy_disabled_tokens:
                logging.warning('Use of deprecated value "%s" for repo '
                                '"%s", layer "%s". Replace with "disabled".',
                                prop, repo_name, lname)
                continue
            elif isinstance(prop, dict):
                layers.append(RepoLayer(name=lname,
                                        repo_path=Path(repo_path),
                                        repo_name=repo_name,
                                        priority=prop.get('prio',
                                                          default_prio)))
            else:
                raise NotImplementedError()
        return layers


class RepoImpl(Repo):
    """
        Provides a generic implementation for a Repo.
    """

    async def fetch_async(self):
        """
            Starts asynchronous repository fetch.
        """

        if self.operations_disabled:
            return

        refdir = get_context().kas_repo_ref_dir
        sdir = os.path.join(refdir, self.qualified_name) if refdir else None

        # fetch to refdir
        if refdir and not os.path.exists(sdir):
            os.makedirs(refdir, exist_ok=True)
            with TemporaryDirectory(prefix=self.qualified_name + '.',
                                    dir=refdir) as tmpdir:
                (retc, _) = await run_cmd_async(
                    self.clone_cmd(tmpdir, createref=True),
                    cwd=get_context().kas_work_dir)

                logging.debug('Created repo ref for %s', self.qualified_name)
                try:
                    os.rename(tmpdir, sdir)
                except OSError:
                    logging.debug('repo %s already cloned by other instance',
                                  self.qualified_name)

        if not os.path.exists(self.path):
            logging.info('Cloning repository %s', self.name)
            os.makedirs(os.path.dirname(self.path), exist_ok=True)
            (retc, _) = await run_cmd_async(
                self.clone_cmd(sdir, createref=False),
                cwd=get_context().kas_work_dir)

        # Make sure the remote origin is set to the value
        # in the kas file to avoid surprises
        try:
            (retc, output) = await run_cmd_async(
                self.set_remote_url_cmd(),
                cwd=self.path)
        except NotImplementedError:
            logging.warning('Repo implementation does not support changing '
                            'the remote url.')

        # take what came out of clone and stick to that forever
        if self.commit is None and self.tag is None and self.branch is None \
           and self.refspec is None:
            return

        # check if we already have the commit. On update, check as well in case
        # the commit is fixed, hence the repo must not be updated anyways
        force_update = get_context().update
        if not force_update or (force_update and self.commit):
            # Do commit/tag/branch/refspec exist in the current repository?
            (retc, output) = await run_cmd_async(self.contains_refspec_cmd(),
                                                 cwd=self.path,
                                                 fail=False)
            if retc == 0:
                logging.info('Repository %s already contains %s as %s',
                             self.name,
                             self.commit or self.tag or self.branch
                             or self.refspec,
                             output.strip())
                # if branch is specified, check if it contains the commit
                # also in our local clone
                depth = get_context().repo_clone_depth
                if self.branch and self.commit and not depth:
                    (_, output) = await run_cmd_async(
                        self.branch_contains_ref(), cwd=self.path, fail=False)
                    if output.strip():
                        return
                else:
                    return

        # Try to fetch if commit/tag/branch/refspec is missing or if --update
        # argument was passed
        (retc, _, err) = await run_cmd_async(self.fetch_cmd(),
                                             cwd=self.path,
                                             fail=False, capture_stderr=True)
        if retc:
            raise RepoFetchError(self, err)
        else:
            logging.info('Repository %s updated', self.name)

    def checkout(self):
        """
            Checks out the correct revision of the repo.
        """
        if self.operations_disabled \
            or (self.commit is None and self.tag is None
                and self.branch is None and self.refspec is None):
            return

        if not get_context().force_checkout:
            # Check if repos is dirty
            if self.dirty:
                logging.warning('Repo %s is dirty - no checkout', self.name)
                return

        if self.tag and self.branch:
            raise RepoRefError(
                f'Both tag "{self.tag}" and branch "{self.branch}" '
                f'cannot be specified for repository "{self.name}"')

        if self.tag:
            (retc, output) = run_cmd(self.resolve_tag_cmd(),
                                     cwd=self.path,
                                     fail=False)
            if retc:
                raise RepoRefError(f'Tag "{self.tag}" cannot be found '
                                   f'in repository "{self.name}"')

            desired_ref = output.strip()
            if self.commit and desired_ref != self.commit:
                if self.revision_annotated_tag != self.commit:
                    # Ensure provided commit and tag match
                    raise RepoRefError(f'Provided tag "{self.tag}" '
                                       f'("{desired_ref}") does not match '
                                       f'provided commit "{self.commit}" in '
                                       f'repository "{self.name}", aborting!')
            is_branch = False
        elif self.branch:
            (retc, output) = run_cmd(self.resolve_branch_cmd(),
                                     cwd=self.path,
                                     fail=False)
            if retc:
                raise RepoRefError(
                    f'Branch "{self.branch}" cannot be found '
                    f'in repository "{self.name}"')
            # check if branch contains the requested commit.
            # skip check on shallow clones, as branch information is missing
            if self.commit and not get_context().repo_clone_depth:
                (_, output) = run_cmd(self.branch_contains_ref(),
                                      cwd=self.path,
                                      fail=False)
                if not output.strip():
                    raise RepoRefError(
                        f'Branch "{self.branch}" in '
                        f'repository "{self.name}" does not contain '
                        f'commit "{self.commit}"')

            desired_ref = self.commit or output.strip()
            is_branch = True
        elif self.commit:
            desired_ref = self.commit
            is_branch = False
        else:
            desired_ref = self.refspec
            is_branch = False

        run_cmd(self.checkout_cmd(desired_ref, is_branch), cwd=self.path)
        logging.info(f'Repository {self.name} checked out to {desired_ref}')

    async def apply_patches_async(self):
        """
            Applies patches to a repository asynchronously.
        """
        if self.operations_disabled or not self._patches:
            return 0

        if self.dirty:
            logging.warning(f'Repo {self.name} is dirty - no patching')
            return 0

        (retc, _) = await run_cmd_async(self.prepare_patches_cmd(),
                                        cwd=self.path)

        my_patches = []

        for patch in self._patches:
            other_repo = get_context().config.repo_dict.get(patch['repo'],
                                                            None)

            if not other_repo:
                raise PatchMappingError('Could not find referenced repo. '
                                        f'(missing repo: {patch["repo"]}, '
                                        f'repo: {self.name}, '
                                        f'patch entry: {patch["id"]})')

            path = os.path.join(other_repo.path, patch['path'])
            cmd = []

            if os.path.isfile(path):
                my_patches.append((path, patch['id']))
            elif os.path.isdir(path) \
                    and os.path.isfile(os.path.join(path, 'series')):
                with open(os.path.join(path, 'series')) as f:
                    for line in f:
                        if line.startswith('#'):
                            continue
                        p = os.path.join(path, line.split(' #')[0].rstrip())
                        if os.path.isfile(p):
                            my_patches.append((p, patch['id']))
                        else:
                            raise PatchFileNotFound(p)
            else:
                raise PatchFileNotFound(
                    'Could not find patch. '
                    f'(patch path: {path}, repo: {self.name}, patch '
                    f'entry: {patch["id"]})')

        for (path, patch_id) in my_patches:
            cmd = self.apply_patches_file_cmd(path)
            (retc, out, err) = await run_cmd_async(
                cmd, cwd=self.path, fail=False, capture_stderr=True)
            if retc:
                raise PatchApplyError(
                    'Could not apply patch. Please fix repos and patches:\n'
                    f'patch path: {path}, repo: {self.name}, patch '
                    f'entry: {patch_id}', cmd, out, err)

            logging.info('Patch applied. '
                         '(patch path: %s, repo: %s, patch entry: %s)',
                         path, self.name, patch_id)

            cmd = self.add_cmd()
            (retc, out, err) = await run_cmd_async(
                cmd, cwd=self.path, fail=False, capture_stderr=True)
            if retc:
                raise PatchApplyError('Could not add patched files: repo: '
                                      f'{self.name}', cmd, out, err)

            timestamp = self.get_patch_timestamp(path)
            if not timestamp:
                dt = datetime.fromtimestamp(os.path.getmtime(path))
                timestamp = dt.astimezone().strftime(
                    "%a, %d %b %Y %H:%M:%S %z")

            env = get_context().environ.copy()
            msg = f'kas: {patch_id}\n\npatch {path} applied by kas'
            cmd = self.commit_cmd(env, 'kas <kas@example.com>', msg,
                                  timestamp)
            (retc, out, err) = await run_cmd_async(
                cmd, cwd=self.path, env=env, fail=False, capture_stderr=True)
            if retc:
                raise PatchApplyError('Could not commit patch changes. repo: '
                                      f'{self.name}', cmd, out, err)

        return 0

    def resolve_local(self):
        (retc, output) = run_cmd(self.get_remote_url_cmd(),
                                 cwd=self.path, fail=False)
        if retc == 0:
            self.url = output.strip()

        (retc, output) = run_cmd(self.get_commit_cmd(),
                                 cwd=self.path, fail=False)
        if retc == 0:
            self.commit = output.strip()
        if self.url and self.commit:
            logging.debug('Repository %s resolved to %s @ %s',
                          self.name, self.url, self.commit)


class GitRepo(RepoImpl):
    """
        Provides the git functionality for a Repo.
    """

    @staticmethod
    def get_type():
        return 'git'

    def remove_ref_prefix(self, ref):
        ref_prefix = 'refs/'
        return ref[ref.startswith(ref_prefix) and len(ref_prefix):]

    def add_cmd(self):
        return ['git', 'add', '-A']

    def clone_cmd(self, srcdir, createref):
        cmd = ['git', 'clone', '-q']

        depth = get_context().repo_clone_depth
        if depth:
            if self.refspec:
                logging.warning('Shallow cloning is not supported for legacy '
                                f'refspec on repository "{self.name}". '
                                'Performing full clone.')
            else:
                if createref:
                    # this is not a user-error, as the clone of the work repo
                    # can still be shallow.
                    logging.debug('Shallow cloning is not supported for '
                                  f'reference repository of "{self.name}". '
                                  'Performing full clone.')
                else:
                    cmd.extend(['--depth', str(depth)])
                if self.branch:
                    cmd.extend(['--branch',
                                self.remove_ref_prefix(self.branch)])

        if createref:
            cmd.extend(['--bare', '--', self.effective_url, srcdir])
        elif srcdir:
            cmd.extend(['--reference', srcdir, '--', srcdir, self.path])
        else:
            cmd.extend(['--', self.effective_url, self.path])
        return cmd

    def commit_cmd(self, env, author, msg, date):
        env["GIT_COMMITTER_DATE"] = date
        return ['git', 'commit', '-a', '--author', author, '-m', msg,
                '--date', date]

    def contains_refspec_cmd(self):
        branch = self.branch or self.refspec
        if branch and branch.startswith('refs/'):
            branch = 'remotes/origin/' + self.remove_ref_prefix(branch)
        return ['git', 'cat-file', '-t', self.commit or self.tag or branch]

    def fetch_cmd(self):
        cmd = ['git', 'fetch', '-q']

        depth = 0 if self.refspec else get_context().repo_clone_depth
        if depth:
            cmd.extend(['--depth', str(depth)])

        if self.tag:
            cmd.extend(['origin', f'+{self.tag}:refs/tags/{self.tag}'])
            return cmd

        # only fetch this commit (branch information is lost)
        if depth and self.commit:
            cmd.extend(['origin', self.commit])
            return cmd

        branch = self.branch or self.refspec
        if branch and (branch.startswith('refs/') or depth):
            branch = self.remove_ref_prefix(branch)
            cmd.extend(['origin', f'+{branch}:refs/remotes/origin/{branch}'])

        return cmd

    def is_dirty_cmd(self):
        return ['git', 'diff', '--stat']

    def is_signed_cmd(self):
        if self.tag:
            return ['git', 'verify-tag', '--raw', self.tag]
        else:
            refspec = self.commit or self.revision
            return ['git', 'verify-commit', '--raw', refspec]

    def resolve_branch_cmd(self):
        refspec = self.remove_ref_prefix(self.branch or self.refspec)
        return ['git', 'rev-parse', '--verify', '-q', f'origin/{refspec}']

    def resolve_tag_cmd(self):
        return ['git', 'rev-list', '-n', '1', self.remove_ref_prefix(self.tag)]

    def resolve_annotated_tag_cmd(self):
        return ['git', 'show-ref', '--tags', '--hash', self.tag]

    def branch_contains_ref(self):
        return ['git', 'branch', f'origin/{self.branch}',
                '-r', '--contains', self.commit]

    def checkout_cmd(self, desired_ref, is_branch):
        desired_ref = self.remove_ref_prefix(desired_ref)
        if re.match(r'^[0-9a-f]{40}|[0-9a-f]{64}$', desired_ref):
            desired_ref += '^{commit}'
        cmd = ['git', 'checkout', '-q', desired_ref]
        if is_branch:
            branch = self.remove_ref_prefix(self.branch or self.refspec)
            branch = branch[branch.startswith('heads/') and len('heads/'):]
            cmd.extend(['-B', branch])
        if get_context().force_checkout:
            cmd.append('--force')
        return cmd

    def prepare_patches_cmd(self):
        refspec = self.commit \
            or self.remove_ref_prefix(self.tag or self.branch or self.refspec)
        return ['git', 'checkout', '-q', '-B', f'patched-{refspec}']

    def apply_patches_file_cmd(self, path):
        return ['git', 'apply', '--whitespace=nowarn', path]

    def set_remote_url_cmd(self):
        return ['git', 'remote', 'set-url', 'origin', self.effective_url]

    def get_remote_url_cmd(self):
        return ['git', 'remote', 'get-url', 'origin']

    def get_commit_cmd(self):
        rev = self.commit or 'HEAD'
        return ['git', 'rev-parse', '--verify', rev]

    def get_patch_timestamp(self, path):
        date = linecache.getline(path, 3)
        linecache.clearcache()
        if date and date.startswith("Date: "):
            return date.replace("Date: ", "").strip()

    def contains_path_cmd(self, path):
        return ['git', 'ls-files', '--error-unmatch', path]

    def diff(self, commit1, commit2):
        if commit1 is None:
            commit1 = 'HEAD'
        if commit2 is None:
            commit2 = 'HEAD'
        git_repo = GitPythonRepo(self.path)
        shallow_file = os.path.join(git_repo.git_dir, 'shallow')
        if os.path.isfile(shallow_file):
            git_repo.git.fetch(unshallow=True)
        try:
            commits = list(git_repo.iter_commits(
                           f'{commit1}..{commit2}'))
        except GitCommandError as e:
            valid1 = git_repo.is_valid_object(commit1)
            valid2 = git_repo.is_valid_object(commit2)
            if not valid1 and not valid2:
                error_msg = f'Commits ({commit1} and {commit2}) are not valid'
            elif not valid1:
                error_msg = f'First commit ({commit1}) is not valid'
            elif not valid2:
                error_msg = f'Second commit ({commit2}) is not valid'
            else:
                error_msg = str(e)
            raise RepoRefError(
                'Could not compute diff for repository '
                f'"{self.name}": {error_msg}')
        diff_json = {self.name: []}
        for commit in commits:
            diff_json[self.name].append({
                'commit': commit.hexsha,
                'author': commit.author.name,
                'email': commit.author.email,
                'commit_date': commit.committed_datetime.
                strftime("%Y-%m-%d %H:%M:%S"),
                'message': commit.message
            })
        return diff_json


class MercurialRepo(RepoImpl):
    """
        Provides the hg functionality for a Repo.
    """

    @staticmethod
    def get_type():
        return 'hg'

    def add_cmd(self):
        return ['hg', 'add']

    def clone_cmd(self, srcdir, createref):
        # Mercurial does not support repo references (object caches)
        if createref:
            return ['true']
        return ['hg', 'clone', self.effective_url, self.path]

    def commit_cmd(self, env, author, msg, date):
        return ['hg', 'commit', '--user', author, '-m', msg, '--date', date]

    def contains_refspec_cmd(self):
        return ['hg', 'log', '-r', self.commit or self.tag or self.branch
                or self.refspec]

    def fetch_cmd(self):
        return ['hg', 'pull']

    def is_dirty_cmd(self):
        return ['hg', 'status', '--modified', '--added',
                '--removed', '--deleted']

    def is_signed_cmd(self):
        raise NotImplementedError()

    def resolve_branch_cmd(self):
        if self.branch:
            return ['hg', 'identify', '--id', '-r',
                    f'limit(heads(branch({self.branch})))']
        else:
            return ['hg', 'identify', '--id', '-r', self.refspec]

    def resolve_tag_cmd(self):
        refspec = self.tag or self.refspec
        return ['hg', 'identify', '--id', '-r', f'tag({refspec})']

    def resolve_annotated_tag_cmd(self):
        raise NotImplementedError("Mercurial does not support annotated tags")

    def branch_contains_ref(self):
        return ['hg', 'log', '-r', self.commit, '-b', self.branch]

    def checkout_cmd(self, desired_ref, is_branch):
        cmd = ['hg', 'checkout', desired_ref]
        if get_context().force_checkout:
            cmd.append('--clean')
        return cmd

    def prepare_patches_cmd(self):
        refspec = (self.commit or self.tag or self.branch or self.refspec)
        # strip revision part from refspec as not allowed in branch names
        refspec = refspec.split(':')[-1]
        return ['hg', 'branch', '-f', f'patched-{refspec}']

    def apply_patches_file_cmd(self, path):
        return ['hg', 'import', '--no-commit', path]

    def set_remote_url_cmd(self):
        raise NotImplementedError()

    def get_remote_url_cmd(self):
        return ['hg', 'paths', 'default']

    def get_commit_cmd(self):
        rev = self.commit or '.'
        return ['hg', 'log', '-r', rev, '--template', '{node}\n']

    def get_patch_timestamp(self, path):
        date = None
        if linecache.getline(path, 3).startswith("# Date "):
            date = linecache.getline(path, 4)
        linecache.clearcache()

        if date and date.startswith("# "):
            return date.replace("# ", "").strip()

    def contains_path_cmd(self, path):
        return ['hg', 'files', path]

    def diff(self, commit1, commit2):
        raise NotImplementedError("Unsupported diff for MercurialRepo")


@dataclass
class RepoLayer:
    """
    Standalone definition of a single bitbake layer.
    """
    name: str
    priority: int
    repo_name: str
    repo_path: Path

    @property
    def path(self):
        return self.repo_path / self.name

    def __lt__(self, other):
        if isinstance(other, RepoLayer):
            # The priority is negated because a higher priority means the
            # element appears earlier in the list.
            return (-self.priority, self.repo_name, self.name) < \
                (-other.priority, other.repo_name, other.name)
        return NotImplemented


class SignatureValidator:
    """
        Handles key loading and signature validation of repos
        based on the configuration.
    """

    @staticmethod
    def import_keys(ctx):
        handler_cfg = {
            'gpg': (GPGKeyHandler,
                    Path(ctx.kas_work_dir) / '.kas_gnupg'),
            'ssh': (SSHKeyHandler,
                    Path(ctx.kas_work_dir) / '.kas_ssh-handler'),
        }
        for name, (handler_cls, dir) in handler_cfg.items():
            signers_cfg = ctx.config.get_signers_config(name)
            if not signers_cfg:
                continue
            dir.mkdir(exist_ok=True)
            dir.chmod(0o700)
            ctx.managed_paths.add(dir)
            ctx.keyhandler[name] = handler_cls(dir, signers_cfg, ctx.config)

        for keyhandler in ctx.keyhandler.values():
            ctx.environ.update(keyhandler.env)

    @staticmethod
    def ensure_valid_if_signed(ctx, repo):
        if not repo.signed:
            return
        valid, keyid = repo.check_signature()
        keyhandler = ctx.keyhandler[repo.signers_type]
        info = keyhandler.get_key_repr(keyid) if keyid else 'No info'
        if valid:
            logging.info(f'Repository {repo.name} signature valid: {info}')
            return
        elif keyid:
            raise RepoRefError(f'Repository {repo.name} is not signed '
                               f'with a trusted key: {info}')

        raise RepoRefError(f'Repository {repo.name} is not signed '
                           'with a trusted key.')



# kas - setup tool for bitbake based projects
#
# Copyright (c) Konsulko Group, 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas for-all-repos`` command.

    When this command is executed, kas will checkout the repositories listed
    in the chosen config file and then execute a specified command in each
    repository. It can be used to query the repository status, automate
    actions such as archiving the layers used in a build or to execute any
    other required commands.

    For example, to print the commit hashes used by each repository used in
    the file ``kas-project.yml`` (assuming they are all git repositories) you
    could run::

        kas for-all-repos kas-project.yml 'git rev-parse HEAD'

    The environment for executing the command in each repository is extended
    to include the following variables:

      * ``KAS_REPO_NAME``: The name of the current repository determined by
        either the name property or by the key used for this repo in the config
        file.

      * ``KAS_REPO_PATH``: The path of the local directory where this
        repository is checked out, relative to the directory where ``kas`` is
        executed.

      * ``KAS_REPO_URL``: The URL from which this repository was cloned, or an
        empty string if no remote URL was given in the config file.

      * ``KAS_REPO_COMMIT``: The commit ID which was checked out for this
        repository, or an empty string if no commit was given in the config
        file.

      * ``KAS_REPO_BRANCH``: The branch which was checked out for this
        repository, or an empty string if no branch was given in the config
        file.

      * ``KAS_REPO_TAG``: The tag which was checked out for this repository,
        or an empty string if no tag was given in the config file.

      * ``KAS_REPO_REFSPEC``: The refspec which was checked out for this
        repository, or an empty string if no refspec was given in the config
        file. This variable is obsolete and will be removed when support for
        refspec keys is removed as well. Migrate your repos to commit/branch
        and use the related variables instead.
"""

import logging
import os
import subprocess
from kas.context import create_global_context
from kas.config import Config
from kas.libcmds import Macro, Command, SetupHome
from kas.libkas import setup_parser_common_args, setup_parser_config_arg
from kas.libkas import setup_parser_keep_config_unchanged_arg
from kas.libkas import setup_parser_preserve_env_arg
from kas.libkas import run_handle_preserve_env_arg
from kas.kasusererror import CommandExecError

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2018'


class ForAllRepos:
    name = 'for-all-repos'
    helpmsg = (
        'Runs a specified command in all checked out repositories.'
    )

    @classmethod
    def setup_parser(cls, parser):
        setup_parser_common_args(parser)
        setup_parser_config_arg(parser)
        setup_parser_preserve_env_arg(parser)
        setup_parser_keep_config_unchanged_arg(parser)
        parser.add_argument('command',
                            help='Command to be executed as a string.')

    def run(self, args):
        ctx = create_global_context(args)
        ctx.config = Config(ctx, args.config)

        run_handle_preserve_env_arg(ctx, os, args, SetupHome)

        macro = Macro()
        macro.add(ForAllReposCommand(args.command))
        macro.run(ctx, args.skip)


class ForAllReposCommand(Command):
    def __init__(self, command):
        super().__init__()
        self.command = command

    def __str__(self):
        return 'for-all-repos'

    def execute(self, ctx):
        for repo in ctx.config.get_repos():
            env = {
                **ctx.environ,
                'KAS_REPO_NAME': repo.name,
                'KAS_REPO_PATH': repo.path,
                'KAS_REPO_URL': '' if repo.operations_disabled else repo.url,
                'KAS_REPO_COMMIT': '' if repo.operations_disabled
                                   else (repo.commit or ''),
                'KAS_REPO_BRANCH': repo.branch or '',
                'KAS_REPO_TAG': repo.tag or '',
                'KAS_REPO_REFSPEC': repo.refspec or '',
            }
            logging.info('%s$ %s', repo.path, self.command)
            retcode = subprocess.call(self.command, shell=True, cwd=repo.path,
                                      env=env)
            if retcode != 0:
                raise CommandExecError([self.command], retcode)


__KAS_PLUGINS__ = [ForAllRepos]



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens, 2025
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
import shutil
import yaml
import gnupg
from pathlib import Path
from subprocess import check_output
from kas import kas
from kas.repos import RepoRefError
from kas.kasusererror import KasUserError
from kas.keyhandler import KeyImportError


@pytest.mark.online
def test_signed_gpg_invalid(monkeykas, tmpdir):
    tdir = tmpdir / 'test_signature_verify'
    shutil.copytree('tests/test_signature_verify', tdir)
    monkeykas.chdir(tdir)
    with pytest.raises(RepoRefError):
        kas.kas(['checkout', 'test-gpg-wrong-key.yml'])


@pytest.mark.online
def test_signed_gpg_remove_key(monkeykas, tmpdir):
    tdir = tmpdir / 'test_signature_verify'
    shutil.copytree('tests/test_signature_verify', tdir)
    monkeykas.chdir(tdir)

    # successfully validate signature
    kas.kas(['checkout', 'test-gpg-key-retention.yml'])

    # change key in config and try again
    with open('test-gpg-key-retention.yml', 'r') as f:
        data = yaml.safe_load(f)
        # use OEs signing key
        data['signers']['jan-kiszka']['fingerprint'] = \
            '2AFB13F28FBBB0D1B9DAF63087EB3D32FB631AD9'
    with open('test-gpg-key-retention.yml', 'w') as f:
        yaml.dump(data, f)
    with pytest.raises(RepoRefError):
        kas.kas(['checkout', 'test-gpg-key-retention.yml'])

    # remove key from config and try again (without deleting the keystore)
    with open('test-gpg-key-retention.yml', 'r') as f:
        data = yaml.safe_load(f)
        del data['signers']
    with open('test-gpg-key-retention.yml', 'w') as f:
        yaml.dump(data, f)
    with pytest.raises(KasUserError):
        kas.kas(['checkout', 'test-gpg-key-retention.yml'])


@pytest.mark.dirsfromenv
@pytest.mark.online
def test_signed_gpg_local_key(monkeykas, tmpdir):
    tdir = tmpdir / 'test_signature_verify'
    shutil.copytree('tests/test_signature_verify', tdir)
    monkeykas.chdir(tdir)
    # we don't want to store keys in this repo, hence download a key,
    # export it and pass to kas
    KEY_FINGERPRINT = 'CA5F8C00F5FBC85466016C808AD4AC6F7AE5E714'
    gnupghome = tdir / 'test_gnupg'
    gnupghome.mkdir()
    gnupghome.chmod(0o700)
    gpg = gnupg.GPG(gnupghome=str(gnupghome))
    gpg.recv_keys('keyserver.ubuntu.com', KEY_FINGERPRINT)
    gpg.export_keys(KEY_FINGERPRINT, armor=True,
                    output=str(tdir / 'jan-kiszka.asc'))

    with pytest.raises(KeyImportError):
        kas.kas(['checkout', 'test-gpg-local-key.yml'])

    # remove key definition with wrong fingerprint
    with open('test-gpg-local-key.yml', 'r') as f:
        data = yaml.safe_load(f)
        del data['signers']['wrong-fp']
    with open('test-gpg-local-key.yml', 'w') as f:
        yaml.dump(data, f)
    kas.kas(['checkout', 'test-gpg-local-key.yml'])


@pytest.mark.dirsfromenv
def test_signed_ssh_key(monkeykas, tmpdir):
    tdir = tmpdir / 'test_signature_verify'
    shutil.copytree('tests/test_signature_verify', tdir)
    monkeykas.chdir(tdir)
    kas_wd = monkeykas.get_kwd()

    repodir = Path(kas_wd / 'testrepo')
    repodir.mkdir()
    # create a new ssh key
    sshdir = Path(tdir / 'ssh')
    sshdir.mkdir()
    sshdir.chmod(0o700)

    monkeykas.chdir(repodir)
    check_output(['ssh-keygen', '-t', 'rsa', '-N', '', '-C', 'Comment Space',
                  '-f', str(sshdir / 'testkey')])
    # create a git repository with a single commit
    check_output(['git', 'init'])
    check_output(['git', 'config', 'user.name', 'kas'])
    check_output(['git', 'config', 'user.email', 'kas@example.com'])
    check_output(['git', 'config', 'gpg.format', 'ssh'])
    check_output(['git', 'config', 'user.signingkey', str(sshdir / 'testkey')])
    check_output(['git', 'commit', '-S', '-m', 'initial commit',
                  '--allow-empty'])
    check_output(['git', 'tag', '-s', 'signed-tag', '-m', 'signed tag'])
    monkeykas.chdir(tdir)

    kas.kas(['checkout', 'test-ssh-key.yml'])

    # now replace the ssh key and check if signature verification fails
    (sshdir / 'testkey').unlink()
    (sshdir / 'testkey.pub').unlink()
    check_output(['ssh-keygen', '-t', 'rsa', '-N', '', '-f',
                  str(sshdir / 'testkey')])

    with pytest.raises(RepoRefError):
        kas.kas(['checkout', 'test-ssh-key.yml'])



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import shutil
import pytest
import re
from kas import kas


def test_build_system(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_build_system')
    shutil.copytree('tests/test_build_system', tdir)
    monkeykas.chdir(tdir)

    kas.kas(['shell', 'test-oe.yml', '-c', 'true'])
    with open('build-env', 'r') as f:
        assert f.readline().strip() == 'openembedded'

    kas.kas(['shell', 'test-isar.yml', '-c', 'true'])
    with open('build-env', 'r') as f:
        assert f.readline().strip() == 'isar'

    kas.kas(['shell', 'test-isar-privileged.yml', '-c', 'true'])
    with open('build-env', 'r') as f:
        assert f.readline().strip() == 'isar'

    kas.kas(['shell', 'test-isar-rootless.yml', '-c', 'true'])
    with open('build-env', 'r') as f:
        assert f.readline().strip() == 'isar'
    with open(monkeykas.get_kbd() / 'conf/local.conf', 'r') as f:
        assert any(re.match(r'^ISAR_ROOTLESS.*"1"', line)
                   for line in f.readlines())

    kas.kas(['shell', 'test-openembedded.yml', '-c', 'true'])
    with open('build-env', 'r') as f:
        assert f.readline().strip() == 'openembedded'


@pytest.mark.dirsfromenv
def test_gitconfig(monkeykas, tmpdir, capsys):
    tdir = str(tmpdir / 'test_gitconfig')
    shutil.copytree('tests/test_build_system', tdir)
    monkeykas.chdir(tdir)

    kas.kas(['shell', 'test-oe.yml', '-c',
             f'git config --get user.name > {tdir}/user.name'])
    with open(f'{tdir}/user.name', 'r') as f:
        assert f.readline().strip() == 'kas User'

    monkeykas.setenv('GITCONFIG_FILE', f'{tdir}/gitconfig')
    with open(f'{tdir}/gitconfig', 'w') as f:
        f.write('[user]\n')
        f.write('\temail = kas@kastest.io\n')
        f.write('[url "git@github.com:"]\n')
        f.write('\tinsteadOf = git://github\n')
        f.write('\tinsteadOf = git://github.io\n')
    kas.kas(['shell', 'test-oe.yml', '-c',
             f'git config --get user.email > {tdir}/user.email'])
    kas.kas(['shell', 'test-oe.yml', '-c',
             'git config --get-all "url.git@github.com:.insteadof" '
             f'> {tdir}/url'])
    # check if user is restored after patching
    with open(f'{tdir}/user.email', 'r') as f:
        assert f.readline().strip() == 'kas@kastest.io'
    # check the multi-key url rewrites
    with open(f'{tdir}/url', 'r') as f:
        assert f.readline().strip() == 'git://github'
        assert f.readline().strip() == 'git://github.io'



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2025
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This plugin implements the ``kas shell`` command.

    When this command is executed, kas will checkout repositories, setup the
    build environment and then start a shell in the build environment. This
    can be used to manually run ``bitbake`` with custom command line options
    or to execute other commands such as ``runqemu``. The ``SHELL`` environment
    variable is inherited from the calling environment and controls which shell
    is started.

    For example, to start a shell in the build environment for the file
    ``kas-project.yml`` you could run::

        kas shell kas-project.yml

    Or to invoke qemu to test an image which has been built::

        kas shell kas-project.yml -c 'runqemu'
"""

import logging
import os
import subprocess
from kas.context import create_global_context
from kas.config import Config
from kas.libcmds import Macro, Command, SetupHome, MakeInteractive
from kas.libkas import setup_parser_common_args, setup_parser_config_arg
from kas.libkas import setup_parser_keep_config_unchanged_arg
from kas.libkas import setup_parser_preserve_env_arg
from kas.libkas import run_handle_preserve_env_arg
from kas.kasusererror import CommandExecError

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2018'

SHELL_HISTORY_FILE = '.kas_shell_history'


class Shell:
    """
        Implements a kas plugin that opens a shell within the kas environment.
    """

    name = 'shell'
    helpmsg = 'Run a shell in the build environment.'

    @classmethod
    def setup_parser(cls, parser):
        """
            Setup the argument parser for the shell plugin
        """

        setup_parser_common_args(parser)
        setup_parser_config_arg(parser)
        setup_parser_preserve_env_arg(parser)
        setup_parser_keep_config_unchanged_arg(parser)
        parser.add_argument('-c', '--command',
                            help='Run command',
                            default='')

    def run(self, args):
        """
            Runs this kas plugin
        """

        ctx = create_global_context(args)
        ctx.config = Config(ctx, args.config)

        run_handle_preserve_env_arg(ctx, os, args, SetupHome)

        macro = Macro()
        macro.add(MakeInteractive())
        macro.add(ShellCommand(args.command))
        macro.run(ctx, args.skip)

    @classmethod
    def get_managed_paths(cls):
        return [SHELL_HISTORY_FILE]


class ShellCommand(Command):
    """
        This class implements the command that starts a shell.
    """

    def __init__(self, cmd):
        super().__init__()
        self.cmd = []
        if cmd:
            self.cmd = cmd

    def __str__(self):
        return 'shell'

    def execute(self, ctx):
        logging.info("To start the default build, run: bitbake -c %s %s",
                     ctx.config.get_bitbake_task(),
                     ' '.join(ctx.config.get_bitbake_targets()))
        cmd = [ctx.environ.get('SHELL', '/bin/sh')]
        if self.cmd:
            cmd.append('-c')
            cmd.append(self.cmd)
        ctx.environ['HISTFILE'] = os.path.join(ctx.kas_work_dir,
                                               SHELL_HISTORY_FILE)
        ret = subprocess.call(cmd, env=ctx.environ, cwd=ctx.build_dir)
        if ret != 0:
            logging.error('Shell returned non-zero exit status')
            raise CommandExecError(cmd, ret, True)


__KAS_PLUGINS__ = [Shell]



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2026
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
    This module provides the ``kas-container-usage`` directive that extracts
    usage information from the ``kas-container`` script and renders it as RST.

    Usage::

        .. kas-container-usage:: synopsis
        .. kas-container-usage:: options
'''
__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2026'

import os
import re
import subprocess

from docutils import nodes
from docutils.parsers.rst import Directive
from sphinx.application import Sphinx
from typing import Dict, Union


KAS_CONTAINER_SCRIPT = os.path.join(
    os.path.dirname(__file__), '..', '..', 'kas-container')


def _get_help_output():
    """Run kas-container --help and return the raw output."""
    result = subprocess.run(
        [KAS_CONTAINER_SCRIPT, '--help'],
        capture_output=True, text=True, check=True)
    return result.stdout


def _parse_help(text):
    """
    Parse the kas-container --help output into structured sections.

    Returns a dict with:
      - 'usage_lines': list of usage line strings
      - 'commands': list of (name, description) tuples
      - 'options': list of (flags, description) tuples
    """
    sections = re.split(
        r'^(Usage:|Positional arguments:|Optional arguments:)\n?',
        text, flags=re.MULTILINE)

    # Build a mapping of section header to content
    parts = {}
    for i in range(1, len(sections), 2):
        parts[sections[i]] = sections[i + 1]

    return {
        'usage_lines': _parse_usage(parts.get('Usage:', '')),
        'commands': _parse_tabbed_entries(parts.get(
            'Positional arguments:', '')),
        'options': _parse_tabbed_entries(parts.get(
            'Optional arguments:', '')),
    }


def _parse_usage(text):
    """Extract usage lines, strip leading 'kas-container' prefix context."""
    return [line.strip() for line in text.splitlines() if line.strip()]


def _parse_tabbed_entries(text):
    """Parse lines of 'key<tab>description' into (key, description) tuples."""
    # Unwrap continuation lines (whitespace-indented lines continuing previous)
    text = re.sub(r'\n\s+\t\t\t', ' ', text)
    entries = []
    for line in text.splitlines():
        match = re.match(r'^([^\t]+?)\t+(.*)$', line)
        if match:
            entries.append((match.group(1), match.group(2)))
    return entries


def _build_synopsis(parsed):
    """Build docutils nodes for the synopsis section."""
    code = nodes.literal_block(text='\n'.join(parsed['usage_lines']))
    code['language'] = 'none'
    return [code]


def _build_options(parsed):
    """Build docutils nodes for the commands and options sections."""
    result = []

    # Definition list for commands
    dlist = nodes.definition_list()
    for name, desc in parsed['commands']:
        item = nodes.definition_list_item()
        term = nodes.term()
        term += nodes.strong(text=name)
        item += term
        defn = nodes.definition()
        defn += nodes.paragraph('', nodes.Text(desc))
        item += defn
        dlist += item
    result.append(dlist)

    return result


def _build_option_list(parsed):
    """Build docutils nodes for the options section."""
    # Option list for options
    olist = nodes.option_list()
    for flags, desc in parsed['options']:
        item = nodes.option_list_item()
        group = nodes.option_group()
        for flag in flags.split(', '):
            opt = nodes.option()
            opt += nodes.option_string(text=flag)
            group += opt
        item += group
        description = nodes.description()
        description += nodes.paragraph('', nodes.Text(desc))
        item += description
        olist += item
    return [olist]


class KasContainerUsageDirective(Directive):
    """
    Directive to include kas-container usage information.

    .. kas-container-usage:: synopsis
    .. kas-container-usage:: options
    """

    required_arguments = 1
    optional_arguments = 0
    has_content = False
    option_spec = {}

    def run(self):
        section = self.arguments[0]
        if section not in ('synopsis', 'commands', 'options'):
            self.state_machine.reporter.error(
                f'Invalid kas-container-usage section: {section}. '
                'Must be "synopsis", "commands", or "options".',
                line=self.lineno)
            return []

        env = self.state.document.settings.env
        env.note_dependency(KAS_CONTAINER_SCRIPT)

        help_output = _get_help_output()
        parsed = _parse_help(help_output)

        if section == 'synopsis':
            return _build_synopsis(parsed)
        elif section == 'commands':
            return _build_options(parsed)
        else:
            return _build_option_list(parsed)


def setup(app: Sphinx) -> Dict[str, Union[bool, str]]:
    app.add_directive('kas-container-usage', KasContainerUsageDirective)

    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
        'version': '1.0',
    }



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2023
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
    This module provides a common base class for all exceptions
    which are related to user or configuration errors. These exceptions
    should be caught and reported to the user using a meaningful message
    instead of a stacktrace.

    When handling errors in KAS, never return directly using `sys.exit`,
    but instead throw an exception derived from :class:`KasUserError` (for user
    errors), or one derived from `Exception` for internal errors. These
    are then handled centrally, mapped to correct return codes and pretty
    printed.
"""


__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2023'


class KasUserError(Exception):
    """
    User or input error. Derive all user error exceptions from this class.
    """
    pass


class CommandExecError(KasUserError):
    """
    Failure in execution of a shell command. The `forward_error_code` parameter
    can be used to request the receiver of the exception to `sys.exit` with
    that code instead of a generic one. Only use this in special cases, where
    the return code can actually be related to a single shell command.
    """
    def __init__(self, command, ret_code,
                 forward_ret_code=False):
        self.ret_code = ret_code
        self.forward = forward_ret_code
        message = ' '.join([f"'{c}'" if ' ' in c else c for c in command])
        super().__init__(f'Command "{message}" failed with error {ret_code}')


class ArgsCombinationError(KasUserError):
    """
    Invalid combination of CLI arguments provided
    """
    def __init__(self, message):
        super().__init__(f'Invalid combination of arguments: {message}')


class ArtifactNotFoundError(KasUserError, FileNotFoundError):
    """
    A configured artifact is not found (or the glob matches 0 elements).
    """
    def __init__(self, name, artifact):
        super().__init__(f'No artifact found for {name}:"{artifact}"')


class EnvSetButNotFoundError(KasUserError, FileNotFoundError):
    """
    A environment variable pointing to a file or directory is set, but
    the path it points to does not exist.
    """
    def __init__(self, env_name, path):
        super().__init__(f'Environment variable "{env_name}" is set, but the '
                         f'path does not exist: {path}')


class MissingModuleError(KasUserError):
    """
    An optional module is missing for the requested operation
    """
    def __init__(self, module, operation) -> None:
        super().__init__(f'Module "{module}" is required for: {operation}')



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2018
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module is the main entry point for kas, setup tool for bitbake based
    projects. In case of user errors (e.g. invalid configuration, repo fetch
    failure) kas exits with error code 2, while exiting with 1 for internal
    errors. When cancelled by SIGINT, kas exits with 130. For details on error
    handling, see :mod:`kas.kasusererror`.
"""

import argparse
import asyncio
import distro
import traceback
import logging
import signal
import sys
import os
from .kasusererror import KasUserError, CommandExecError

try:
    import colorlog
    HAVE_COLORLOG = True
except ImportError:
    HAVE_COLORLOG = False

from . import __version__, __file_version__, __compatible_file_version__
from . import plugins

__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2018'

DEFAULT_LOG_LEVEL = 'info'


def create_logger():
    """
        Setup the logging environment
    """
    log = logging.getLogger()  # root logger
    set_global_loglevel(DEFAULT_LOG_LEVEL.upper())
    format_str = '%(asctime)s - %(levelname)-8s - %(message)s'
    date_format = '%Y-%m-%d %H:%M:%S'
    if HAVE_COLORLOG and os.isatty(2):
        cformat = '%(log_color)s' + format_str
        colors = {'DEBUG': 'reset',
                  'INFO': 'reset',
                  'WARNING': 'bold_yellow',
                  'ERROR': 'bold_red',
                  'CRITICAL': 'bold_red'}
        formatter = colorlog.ColoredFormatter(cformat, date_format,
                                              log_colors=colors)
    else:
        formatter = logging.Formatter(format_str, date_format)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    log.addHandler(stream_handler)
    return logging.getLogger(__name__)


def cleanup_logger():
    """
        Cleanup the logging environment
    """
    for handler in logging.root.handlers[:]:
        if isinstance(handler, logging.StreamHandler):
            logging.root.removeHandler(handler)


def set_global_loglevel(level):
    """
        Configure the global log level.
        Implemented as function to monkey-patch it out in tests.
    """
    logging.getLogger().setLevel(level)


def register_signal_handlers(loop):
    """
        Register the signal handlers which should be handled by the
        event loop. Implemented as function to monkey-patch it out in tests.
    """
    loop.add_signal_handler(signal.SIGTERM, interruption)
    loop.add_signal_handler(signal.SIGINT, interruption)


def interruption():
    """
        Gracefully cancel all tasks in the event loop
    """
    loop = asyncio.get_event_loop()
    for sig in [signal.SIGINT, signal.SIGTERM]:
        loop.remove_signal_handler(sig)
        loop.add_signal_handler(sig, termination)
    pending = asyncio.all_tasks(loop)
    if pending:
        logging.debug(f'waiting for {len(pending)} tasks to terminate')
    [t.cancel() for t in pending]


def termination():
    """
        Forcefully terminate the process
    """
    logging.error('kas terminated forcefully')
    os._exit(130)


def shutdown_loop(loop):
    """
        Waits for completion of the event loop but ignores any exceptions.
        The tasks are either already cancelled or will be transitively
        cancelled shortly. As this is the final cleanup, we cannot check for
        exceptions as these might lead in an unclosed event loops.
    """
    pending = asyncio.all_tasks(loop)
    if len(pending):
        logging.debug("Cleanup %d remaining tasks", len(pending))
    loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))
    loop.close()


def kas_get_argparser():
    """
        Creates an argparser for kas with all plugins.
    """

    # Load plugins here so that the commands and arguments introduced by the
    # plugins can be seen by sphinx when it calls this function to build the
    # documentation
    plugins.load()

    parser = argparse.ArgumentParser(description='kas - setup tool for '
                                     'bitbake based project')

    verstr = f'%(prog)s {__version__} ' \
             f'(configuration format version {__file_version__}, ' \
             f'earliest compatible version {__compatible_file_version__})'
    parser.add_argument('--version', action='version', version=verstr)

    parser.add_argument('-l', '--log-level',
                        choices=['debug', 'info', 'warning', 'error',
                                 'critical'],
                        default=f'{DEFAULT_LOG_LEVEL}',
                        help=f'Set log level (default: {DEFAULT_LOG_LEVEL})')

    subparser = parser.add_subparsers(dest='cmd')

    for plugin in plugins.all():
        plugin_parser = subparser.add_parser(
            plugin.name,
            help=plugin.helpmsg,
            formatter_class=ArgumentChoicesHelpFormatter)
        plugin.setup_parser(plugin_parser)

    return parser


class ArgumentChoicesHelpFormatter(argparse.HelpFormatter):
    """Help message formatter which adds choices to argument help.

    If the default METAVAR is used, this will do nothing, as the default
    METAVAR shows the available choices already. If the METAVAR is
    overridden, and %(choice)s is not present in the help string, add
    them.
    """

    def _get_help_string(self, action):
        help = action.help
        if action.choices and action.metavar is not None:
            if "%(choices)" not in action.help:
                help += " Possible choices: %(choices)s."
        return help


def kas(argv):
    """
        The actual main entry point of kas.
    """
    create_logger()

    parser = kas_get_argparser()
    args = parser.parse_args(argv)

    if args.log_level:
        set_global_loglevel(args.log_level.upper())

    logging.info('%s %s started on %s %s', os.path.basename(sys.argv[0]),
                 __version__, distro.name(),
                 distro.codename() or distro.version())

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    register_signal_handlers(loop)

    try:
        plugin_class = plugins.get(args.cmd)
        if plugin_class:
            plugin = plugin_class()
            plugin.run(args)
        else:
            parser.print_help()
    except CommandExecError as err:
        logging.error('%s', err)
        raise
    except KasUserError as err:
        logging.error('%s', err)
        raise
    except asyncio.CancelledError:
        logging.error('kas execution cancelled')
        raise
    except Exception as err:
        logging.error('%s', err)
        raise
    finally:
        shutdown_loop(loop)
        cleanup_logger()


def main():
    """
        The main function that operates as a wrapper around kas.
    """

    try:
        kas(sys.argv[1:])
    except CommandExecError as err:
        sys.exit(err.ret_code if err.forward else 2)
    except KasUserError:
        sys.exit(2)
    except asyncio.CancelledError:
        sys.exit(130)
    except Exception:
        traceback.print_exc()
        sys.exit(1)


if __name__ == '__main__':
    main()



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import shutil
import pytest
import json
from kas import kas
from kas.kasusererror import ArtifactNotFoundError

BITBAKE_OPTIONS_SHA256 = "e35d535e81cfdc4ed304af8000284c36" \
                         "19d2c4c78392ddcefe9ca46b158235f8"


@pytest.mark.dirsfromenv
def test_artifact_node(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_build')
    shutil.copytree('tests/test_build', tdir)
    monkeykas.chdir(tdir)
    kas.kas(['build', 'artifact-named.yml'])
    kas.kas(['build', 'artifact-glob.yml'])
    kas.kas(['build', 'artifact-invalid.yml'])


@pytest.mark.online
def test_provenance(monkeykas, tmpdir):
    tdir = str(tmpdir / 'test_build')
    shutil.copytree('tests/test_build', tdir)
    monkeykas.chdir(tdir)
    kas_bd = monkeykas.get_kbd()

    with pytest.raises(ArtifactNotFoundError):
        kas.kas(['build', '--provenance', 'mode=min',
                 'artifact-invalid.yml'])

    provenance_path = kas_bd / 'attestation/kas-build.provenance.json'
    kas.kas(['build', '--provenance', 'mode=min', 'provenance.yml'])
    with open(provenance_path, 'r') as f:
        prov = json.load(f)
        assert prov['subject'][0]['name'] == 'bitbake.options'
        assert 'env' not in \
            prov['predicate']['buildDefinition']['internalParameters']

    with monkeykas.context() as mp:
        mp.setenv('CAPTURE_THIS', 'OK Sir!')
        kas.kas(['build', '--provenance', 'mode=max', 'provenance.yml'])
    with open(provenance_path, 'r') as f:
        prov = json.load(f)
        params = prov['predicate']['buildDefinition']['internalParameters']
        assert params['env']['CAPTURE_THIS'] == 'OK Sir!'
        assert prov['subject'][0]['name'] == 'bitbake.options'
        assert prov['subject'][0]['digest']['sha256'] == BITBAKE_OPTIONS_SHA256



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2018
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import os
import io
import textwrap
import contextlib

import pytest

from kas import includehandler, context
from kas.includehandler import ConfigFile


@pytest.fixture(autouse=True)
def fixed_version(monkeypatch):
    monkeypatch.setattr(includehandler, '__file_version__', 5)
    monkeypatch.setattr(includehandler, '__compatible_file_version__', 4)


@pytest.fixture(autouse=True)
def with_kas_context():
    context.create_global_context(None)
    yield
    context.__context__ = None


class MockFileIO(io.StringIO):
    def close(self):
        self.seek(0)


def mock_file(indented_content):
    return MockFileIO(textwrap.dedent(indented_content))


@contextlib.contextmanager
def patch_open(component, string='', dictionary=None):
    dictionary = dictionary or {}
    old_attr = getattr(component, 'open', None)
    component.open = lambda f, *a, **k: mock_file(dictionary.get(f, string))
    yield
    if old_attr:
        component.open = old_attr
    else:
        del component.open


class TestLoadConfig:
    def test_err_invalid_ext(self):
        # Test for invalid file extension:
        exception = includehandler.LoadConfigException
        with pytest.raises(exception):
            ConfigFile.load('x.xyz')

    def util_exception_content(self, testvector):
        for string, exception in testvector:
            with patch_open(includehandler, string=string):
                with pytest.raises(exception):
                    ConfigFile.load('x.yml')

    def test_err_header_missing(self):
        exception = includehandler.LoadConfigException
        testvector = [
            ('', exception),
            ('a', exception),
            ('1', exception),
            ('a:', exception)
        ]

        self.util_exception_content(testvector)

    def test_err_header_invalid_type(self):
        exception = includehandler.LoadConfigException
        testvector = [
            ('header:', exception),
            ('header: 1', exception),
            ('header: a', exception),
            ('header: []', exception),
        ]

        self.util_exception_content(testvector)

    def test_err_version_missing(self):
        exception = includehandler.LoadConfigException
        testvector = [
            ('header: {}', exception),
            ('header: {a: 1}', exception),
        ]

        self.util_exception_content(testvector)

    def test_err_version_invalid_format(self):
        exception = includehandler.LoadConfigException
        testvector = [
            ('header: {version: "0.5"}', exception),
            ('header: {version: "x"}', exception),
            ('header: {version: 3}', exception),
            ('header: {version: 6}', exception),
        ]

        self.util_exception_content(testvector)

    def test_err_parse_yaml(self):
        exception = includehandler.LoadConfigException
        testvector = [
            # misaligned column
            ('header:\n  version: 17\n repo:', exception),
        ]
        self.util_exception_content(testvector)

    def test_header_valid(self):
        testvector = [
            'header: {version: 4}',
            'header: {version: 5}',
        ]
        for string in testvector:
            with patch_open(includehandler, string=string):
                ConfigFile.load('x.yml')

    def test_compat_version(self, monkeypatch):
        monkeypatch.setattr(includehandler, '__compatible_file_version__', 1)
        with patch_open(includehandler, string='header: {version: "0.10"}'):
            ConfigFile.load('x.yml')

    def test_source_dir_rejected_in_non_main_file(self):
        string = 'header: {version: 5}\n_source_dir: /some/path'
        with patch_open(includehandler, string=string):
            with pytest.raises(includehandler.LoadConfigException):
                ConfigFile.load('x.yml', is_main_file=False)

    def test_source_dir_allowed_in_main_file(self):
        string = 'header: {version: 5}\n_source_dir: /some/path'
        with patch_open(includehandler, string=string):
            cf = ConfigFile.load('x.yml', is_main_file=True)
            assert cf.src_dir == '/some/path'

    def test_signers_require_path_or_keyserver(self):
        exception = includehandler.LoadConfigException
        testvector = [
            # signer with neither path nor gpg_keyserver
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    type: gpg\n', exception),
        ]
        self.util_exception_content(testvector)

    def test_signers_path_requires_repo(self):
        exception = includehandler.LoadConfigException
        testvector = [
            # path without repo
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    path: key.asc\n', exception),
        ]
        self.util_exception_content(testvector)

    def test_signers_keyserver_requires_fingerprint(self):
        exception = includehandler.LoadConfigException
        testvector = [
            # gpg_keyserver without fingerprint
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    gpg_keyserver: keyserver.ubuntu.com\n', exception),
        ]
        self.util_exception_content(testvector)

    def test_signers_ssh_requires_path_and_repo(self):
        exception = includehandler.LoadConfigException
        testvector = [
            # ssh type with keyserver but no path
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    type: ssh\n'
             '    gpg_keyserver: keyserver.ubuntu.com\n'
             '    fingerprint: 2AFB13F28FBBB0D1B9DAF63087EB3D32FB631AD9\n',
             exception),
            # ssh type with path but no repo
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    type: ssh\n'
             '    path: key.pub\n',
             exception),
        ]
        self.util_exception_content(testvector)

    def test_signers_valid(self):
        testvector = [
            # path with repo
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    repo: this\n'
             '    path: key.asc\n'),
            # keyserver with fingerprint
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    gpg_keyserver: keyserver.ubuntu.com\n'
             '    fingerprint: 2AFB13F28FBBB0D1B9DAF63087EB3D32FB631AD9\n'),
            # ssh type with path and repo
            ('header: {version: 5}\n'
             'signers:\n'
             '  mysigner:\n'
             '    type: ssh\n'
             '    repo: this\n'
             '    path: key.pub\n'),
        ]
        for string in testvector:
            with patch_open(includehandler, string=string):
                ConfigFile.load('x.yml')


class TestIncludes:
    header = '''
header:
  version: 5
{}'''

    def util_include_content(self, testvector, monkeypatch):
        # disable schema validation for these tests:
        monkeypatch.setattr(includehandler, 'CONFIGSCHEMA', {})
        for test in testvector:
            with patch_open(includehandler, dictionary=test['fdict']):
                ginc = includehandler.IncludeHandler(['x.yml'])
                config, missing = ginc.get_config(repos=test['rdict'])

                # Remove header, because we dont want to compare it:
                config.pop('header')

                assert test['conf'] == config
                assert test['rmiss'] == missing

                if 'cfiles' in test:
                    assert len(ginc.config_files) == len(test['cfiles'])

                    for i, cf in enumerate(ginc.config_files):
                        assert test['cfiles'][i][0] == str(cf.filename)
                        assert test['cfiles'][i][1] == cf.is_lockfile
                        assert test['cfiles'][i][2] == cf.is_external

    def test_valid_includes_none(self, monkeypatch):
        header = self.__class__.header
        testvector = [
            {
                'fdict': {
                    'x.yml': header.format('')
                },
                'rdict': {
                },
                'conf': {
                },
                'rmiss': [
                ],
                'cfiles': [
                    (
                        'x.yml',
                        False,
                        False
                    )
                ]
            },
        ]

        self.util_include_content(testvector, monkeypatch)

    def test_valid_includes_some(self, monkeypatch):
        header = self.__class__.header
        testvector = [
            # Include one file from the same repo:
            {
                'fdict': {
                    'x.yml': header.format('  includes: ["y.yml"]'),
                    os.path.abspath('y.yml'): header.format('\nv:')
                },
                'rdict': {
                },
                'conf': {
                    'v': None
                },
                'rmiss': [
                ],
                'cfiles': [
                    (
                        os.path.abspath('y.yml'),
                        False,
                        False
                    ),
                    (
                        'x.yml',
                        False,
                        False
                    )
                ]
            },
            # Include one file from another not available repo:
            {
                'fdict': {
                    'x.yml': header.format(
                        '  includes: [{repo: rep, file: y.yml}]'),
                },
                'rdict': {
                },
                'conf': {
                },
                'rmiss': [
                    'rep',
                ],
                'cfiles': [
                    (
                        'x.yml',
                        False,
                        False
                    )
                ]
            },
            # Include one file from the same repo and one from another
            # not available repo:
            {
                'fdict': {
                    'x.yml': header.format('  includes: ["y.yml", '
                                           '{repo: rep, file: y.yml}]'),
                    os.path.abspath('y.yml'): header.format('\nv:')
                },
                'rdict': {
                },
                'conf': {
                    'v': None
                },
                'rmiss': [
                    'rep',
                ],
                'cfiles': [
                    (
                        os.path.abspath('y.yml'),
                        False,
                        False
                    ),
                    (
                        'x.yml',
                        False,
                        False
                    )
                ]
            },
            # Include one file from another available repo:
            {
                'fdict': {
                    'x.yml': header.format(
                        '  includes: [{repo: rep, file: y.yml}]'),
                    '/rep/y.yml': header.format('\nv:')
                },
                'rdict': {
                    'rep': '/rep'
                },
                'conf': {
                    'v': None
                },
                'rmiss': [
                ],
                'cfiles': [
                    (
                        '/rep/y.yml',
                        False,
                        True
                    ),
                    (
                        'x.yml',
                        False,
                        False
                    )
                ]
            },
            # Include two files from another repo in sub-directories:
            {
                'fdict': {
                    'x.yml': header.format(
                        '  includes: [{repo: rep, file: dir1/y.yml}]'),
                    '/rep/dir1/y.yml': header.format(
                        '  includes: ["dir2/z.yml"]'),
                    '/rep/dir2/z.yml': header.format('\nv:')
                },
                'rdict': {
                    'rep': '/rep'
                },
                'conf': {
                    'v': None
                },
                'rmiss': [
                ],
                'cfiles': [
                    (
                        '/rep/dir2/z.yml',
                        False,
                        True
                    ),
                    (
                        '/rep/dir1/y.yml',
                        False,
                        True
                    ),
                    (
                        'x.yml',
                        False,
                        False
                    )
                ]
            },
        ]

        self.util_include_content(testvector, monkeypatch)

    def test_valid_overwriting(self, monkeypatch):
        header = self.__class__.header
        testvector = [
            {
                'fdict': {
                    'x.yml': header.format('''  includes: ["y.yml"]
v: x'''),
                    os.path.abspath('y.yml'): header.format('''
v: y''')
                },
                'rdict': {
                },
                'conf': {
                    'v': 'x'
                },
                'rmiss': [
                ]
            },
            {
                'fdict': {
                    'x.yml': header.format('''  includes: ["y.yml"]
v: {v: x}'''),
                    os.path.abspath('y.yml'): header.format('''
v: {v: y}''')
                },
                'rdict': {
                },
                'conf': {
                    'v': {'v': 'x'}
                },
                'rmiss': [
                ]
            },
            {
                'fdict': {
                    'x.yml': header.format('''  includes: ["y.yml"]
v1:
v2: []
v3:
  - a: c'''),
                    os.path.abspath('y.yml'): header.format('''
v1: a
v2: [a]
v3:
  - a: b
  - d: c}]''')
                },
                'rdict': {
                },
                'conf': {
                    'v1': None,
                    'v2': [],
                    'v3': [{'a': 'c'}]
                },
                'rmiss': [
                ]
            },
        ]

        self.util_include_content(testvector, monkeypatch)

    def test_valid_merging(self, monkeypatch):
        header = self.__class__.header
        testvector = [
            {
                'fdict': {
                    'x.yml': header.format('''  includes: ["y.yml"]
v1: x
v3:
  a: b
  b:
    e:
  c: d'''),
                    os.path.abspath('y.yml'): header.format('''
v2: y
v3:
  d: e
  b:
    c:
  e: f''')
                },
                'rdict': {
                },
                'conf': {
                    'v1': 'x',
                    'v2': 'y',
                    'v3': {
                        'a': 'b',
                        'b': {'c': None, 'e': None},
                        'c': 'd',
                        'd': 'e',
                        'e': 'f'}
                },
                'rmiss': [
                ]
            },
        ]

        self.util_include_content(testvector, monkeypatch)

    def test_valid_ordering(self, monkeypatch):
        # disable schema validation for this test:
        monkeypatch.setattr(includehandler, 'CONFIGSCHEMA', {})
        header = self.__class__.header
        data = {'x.yml':
                header.format('''  includes: ["y.yml", "z.yml"]
v: {v1: x, v2: x}'''),
                os.path.abspath('y.yml'):
                header.format('''  includes: ["z.yml"]
v: {v2: y, v3: y, v5: y}'''),
                os.path.abspath('z.yml'): header.format('''
v: {v3: z, v4: z}''')}
        with patch_open(includehandler, dictionary=data):
            ginc = includehandler.IncludeHandler(['x.yml'])
            config, _ = ginc.get_config()
            keys = list(config['v'].keys())
            index = {keys[i]: i for i in range(len(keys))}

            # Check for vars in z.yml:
            assert index['v3'] < index['v1']
            assert index['v3'] < index['v2']
            assert index['v3'] < index['v5']
            assert index['v4'] < index['v1']
            assert index['v4'] < index['v2']
            assert index['v4'] < index['v5']

            # Check for vars in y.yml:
            assert index['v2'] < index['v1']
            assert index['v3'] < index['v1']
            assert index['v5'] < index['v1']

    def test_err_string_include_traversal(self, monkeypatch):
        monkeypatch.setattr(includehandler, 'CONFIGSCHEMA', {})
        header = self.__class__.header
        fdict = {
            'x.yml': header.format(
                '  includes: [{repo: rep, file: y.yml}]'),
            '/top-repo/repos/rep/y.yml': header.format(
                '  includes: ["../../other-repo/secret.yml"]'),
        }
        with patch_open(includehandler, dictionary=fdict):
            ginc = includehandler.IncludeHandler(['x.yml'])
            with pytest.raises(includehandler.IncludeException):
                ginc.get_config(repos={'rep': '/top-repo/repos/rep'})

    def test_err_repo_include_traversal(self, monkeypatch):
        monkeypatch.setattr(includehandler, 'CONFIGSCHEMA', {})
        header = self.__class__.header
        fdict = {
            'x.yml': header.format(
                '  includes: [{repo: rep, '
                'file: "../../other-repo/secret.yml"}]'),
        }
        with patch_open(includehandler, dictionary=fdict):
            ginc = includehandler.IncludeHandler(['x.yml'])
            with pytest.raises(includehandler.IncludeException):
                ginc.get_config(repos={'rep': '/top-repo/repos/rep'})



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2020
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import shutil
from kas import kas

import pytest

LAYERBASE = '${TOPDIR}/..'


class DoKas:
    def __init__(self, monkeykas, tmpdir):
        self.tmpdir = tmpdir
        self.monkeykas = monkeykas

    def run(self, filename):
        tdir = str(self.tmpdir / 'test_layers')
        shutil.copytree('tests/test_layers', tdir)
        self.monkeykas.chdir(tdir)
        self.monkeykas.setenv('KAS_CLONE_DEPTH', '1')
        kas.kas(['shell', filename, '-c', 'true'])
        # extract layer path from bblayers, keep order
        with open(self.monkeykas.get_kbd() / 'conf/bblayers.conf', 'r') as f:
            return [x.strip(' \\"\n').replace(LAYERBASE, '')
                    for x in f.readlines() if x.lstrip().startswith(LAYERBASE)]


@pytest.fixture
def dokas(monkeykas, tmpdir):
    return DoKas(monkeykas, tmpdir)


@pytest.mark.online
def test_layers_default(dokas):
    layers = dokas.run('test.yml')
    assert len([l_ for l_ in layers if l_ == '/kas']) == 1


@pytest.mark.online
def test_layers_include(dokas):
    layers = dokas.run('test.yml')
    assert len([l_ for l_ in layers if '/kas1/meta-' in l_]) == 2


@pytest.mark.online
def test_layers_exclude(dokas):
    layers = dokas.run('test.yml')
    assert not any([l_ for l_ in layers if '/kas2' in l_])


@pytest.mark.online
def test_layers_strip_dot(dokas):
    layers = dokas.run('test.yml')
    assert any([l_ for l_ in layers if l_ == '/kas3'])
    assert any([l_ for l_ in layers if l_ == '/kas3/meta-bar'])


@pytest.mark.online
def test_layers_order(dokas):
    layers = dokas.run('test.yml')
    # layers of a repo are sorted alphabetically
    assert layers[1] == '/kas1/meta-bar'
    assert layers[2] == '/kas1/meta-foo'
    # repos are sorted alphabetically (aa-kas from kas4 is last)
    assert layers[-1] == '/aa-kas/meta'


@pytest.mark.online
def test_layers_prio(dokas, monkeykas):
    layers = dokas.run('test-layer-prio.yml')
    # layers are sorted by global priority
    # highest prio (10)
    assert layers[0] == '/02-kas/meta-foo'
    # no prio, sorted alphabetically by repo name, layer name
    assert layers[1] == '/01-kas/aa-test'
    assert layers[2] == '/01-kas/zz-test'
    # default prio as not explicitly specified, sorted by repo name
    assert layers[3] == ''
    # lower than default prio (-10)
    assert layers[4] == '/01-kas'
    # even lower (-20)
    assert layers[5] == '/02-kas/meta-bar'



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2026
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
from kas.attestation import Provenance


@pytest.mark.parametrize('url,expected', [
    # plain https
    ('https://example.com/siemens/kas.git',
     'https://example.com/siemens/kas.git'),
    # https with user:password
    ('https://user:password@example.com/siemens/kas.git',
     'https://example.com/siemens/kas.git'),
    # https with user only
    ('https://user@example.com/siemens/kas.git',
     'https://example.com/siemens/kas.git'),
    # https with port and credentials
    ('https://user:password@example.com:8443/siemens/kas.git',
     'https://example.com:8443/siemens/kas.git'),
    # http with credentials
    ('http://token:x-oauth-basic@example.com/repo.git',
     'http://example.com/repo.git'),
    # ssh without credentials
    ('ssh://git@example.com/siemens/kas.git',
     'ssh://example.com/siemens/kas.git'),
    # plain URL without credentials
    ('https://example.com/path?query=1#frag',
     'https://example.com/path?query=1#frag'),
    # empty string
    ('', ''),
])
def test_strip_credentials(url, expected):
    assert Provenance._strip_credentials(url) == expected



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from pathlib import Path
import shutil
import subprocess
import yaml
import pytest
from kas import kas
from kas.includehandler import IncludeException


def test_transitive_includes(monkeykas, tmpdir, capsys):
    """
    Check if a reference to a repo that itself is specified in a
    remote repo is correctly resolved.
    """
    repos = ['main', 'foo', 'bar']
    dirs = {}
    for r in repos:
        dirs[r] = Path(tmpdir / r)
        os.mkdir(dirs[r])

    # create main repo config
    with open('tests/test_transitive_includes/main.yml', 'rb') as fds:
        config_top = yaml.safe_load(fds)
    config_top['repos']['foo']['url'] = str(dirs['foo'])
    with open(dirs['main'] / 'main.yml', 'w') as config_file:
        yaml.dump(config_top, config_file)
    shutil.copy('tests/test_transitive_includes/oe-init-build-env',
                dirs['main'])

    # create foo repo and config
    with open('tests/test_transitive_includes/foo.yml', 'rb') as fds:
        config_foo = yaml.safe_load(fds)
    config_foo['repos']['bar']['url'] = str(dirs['bar'])
    with open(dirs['foo'] / 'foo.yml', 'w') as config_file:
        yaml.dump(config_foo, config_file)
    subprocess.check_call(['git', 'init'], cwd=dirs['foo'])
    subprocess.check_call(['git', 'config', 'user.email', 'test'],
                          cwd=dirs['foo'])
    subprocess.check_call(['git', 'config', 'user.name', 'test'],
                          cwd=dirs['foo'])
    subprocess.check_call(['git', 'checkout', '-b', 'kas'], cwd=dirs['foo'])
    subprocess.check_call(['git', 'add', 'foo.yml'], cwd=dirs['foo'])
    subprocess.check_call(['git', 'commit', '-m', 'init'], cwd=dirs['foo'])

    # create bar repo
    shutil.copy('tests/test_transitive_includes/bar.yml', dirs['bar'])
    subprocess.check_call(['git', 'init'], cwd=dirs['bar'])
    subprocess.check_call(['git', 'config', 'user.email', 'test'],
                          cwd=dirs['bar'])
    subprocess.check_call(['git', 'config', 'user.name', 'test'],
                          cwd=dirs['bar'])
    subprocess.check_call(['git', 'checkout', '-b', 'kas'], cwd=dirs['bar'])
    subprocess.check_call(['git', 'add', 'bar.yml'], cwd=dirs['bar'])
    subprocess.check_call(['git', 'commit', '-m', 'init'], cwd=dirs['bar'])

    monkeykas.chdir(dirs['main'])
    kas.kas(['dump', 'main.yml'])
    config_final = yaml.safe_load(capsys.readouterr().out)
    assert config_final['env']['KAS_REPO_BAR'] == "1"


@pytest.mark.dirsfromenv
def test_include_from_submodule(monkeykas, tmpdir):
    tmpdir.mkdir('main')
    tmpdir.mkdir('sub')
    for s, t in [('sub_repo.yml', 'sub'),
                 ('super_repo.yml', 'main'),
                 ('oe-init-build-env', 'main')]:
        shutil.copy(f'tests/test_transitive_includes/{s}', tmpdir / t)

    monkeykas.chdir(tmpdir / 'sub')
    subprocess.check_call(['git', 'init', '-b', 'main'])
    subprocess.check_call(['git', 'add', 'sub_repo.yml'])
    subprocess.check_call(['git', 'commit', '-m', 'init'])

    monkeykas.chdir(tmpdir / 'main')
    subprocess.check_call(['git', 'init', '-b', 'main'])

    # clone a repo as non-submodule
    subprocess.check_call(['git', 'clone', '../sub', 'sub'])
    with pytest.raises(IncludeException):
        kas.kas(['checkout', 'super_repo.yml:sub/sub_repo.yml'])

    # now add the repo as submodule
    shutil.rmtree('sub')
    subprocess.check_call(['git', '-c', 'protocol.file.allow=always',
                          'submodule', 'add', '../sub', 'sub'])
    kas.kas(['-l', 'debug', 'checkout', 'super_repo.yml:sub/sub_repo.yml'])



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2018
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
    This module contains the implementation of the kas context.
"""

import distro
import os
import logging
from enum import Enum
from kas.kasusererror import KasUserError
from kas import __version__

__context__ = None


def get_distro_id_base():
    """
        Returns a compatible distro id.
    """
    return distro.like() or distro.id()


def create_global_context(args):
    """
        Creates global context as singleton.
    """
    global __context__
    __context__ = Context(args)
    return __context__


def get_context():
    """
        Returns singleton global context.
    """
    return __context__


class ManagedEnvironment(Enum):
    """
    Managed environments are well-known executors (like CI systems)
    that kas can detect and adapt to.
    """
    GITHUB_ACTIONS = 1
    GITLAB_CI = 2
    VSCODE_REMOTE_CONTAINERS = 3

    def __str__(self):
        if self == self.GITHUB_ACTIONS:
            return 'GitHub Actions'
        if self == self.GITLAB_CI:
            return 'GitLab CI'
        if self == self.VSCODE_REMOTE_CONTAINERS:
            return 'VSCode Remote Containers'
        return f'{self.name}'


class Context:
    """
        Implements the kas build context.
    """
    def __init__(self, args):
        work_dir = os.environ.get('KAS_WORK_DIR', os.getcwd())
        self.__kas_work_dir = os.path.abspath(work_dir)
        build_dir = os.environ.get('KAS_BUILD_DIR',
                                   os.path.join(self.__kas_work_dir, 'build'))
        self.__kas_build_dir = os.path.abspath(build_dir)
        ref_dir = os.environ.get('KAS_REPO_REF_DIR', None)
        self.__kas_repo_ref_dir = os.path.abspath(ref_dir) if ref_dir else None
        clone_depth = os.environ.get('KAS_CLONE_DEPTH', '0')
        if not clone_depth.isdigit():
            raise KasUserError('KAS_CLONE_DEPTH must be a number')
        self.repo_clone_depth = max(int(clone_depth), 0)
        self.setup_initial_environ()
        self.check_container_call()
        # Register the paths that kas created and exclusively owns
        self.managed_paths = set()
        if not os.environ.get('KAS_BUILD_DIR'):
            self.managed_paths.add(self.__kas_build_dir)
        self.keyhandler = {}
        self.config = None
        self.args = args
        self.unpinned_repo_warnings = True

    def setup_initial_environ(self):
        """
            Sets the environment variables for processes that are
            started by kas.
        """
        self.environ = {}
        distro_bases = get_distro_id_base().lower().split()
        for distro_base in distro_bases:
            if distro_base in ['fedora', 'suse', 'opensuse']:
                self.environ = {'LC_ALL': 'en_US.utf8',
                                'LANG': 'en_US.utf8',
                                'LANGUAGE': 'en_US'}
                break
            elif distro_base in ['debian', 'ubuntu', 'gentoo', 'arch']:
                self.environ = {'LC_ALL': 'en_US.UTF-8',
                                'LANG': 'en_US.UTF-8',
                                'LANGUAGE': 'en_US:en'}
                break
        if self.environ == {}:
            logging.warning('kas: No supported distros found in %s. '
                            'No default locales set.', distro_bases)

        for key in ['http_proxy', 'https_proxy', 'ftp_proxy', 'no_proxy',
                    'SSH_AUTH_SOCK',
                    'BB_NUMBER_THREADS', 'PARALLEL_MAKE']:
            val = os.environ.get(key, None)
            if val:
                self.environ[key] = val

        # make remote containers environment available in kas
        if self.managed_env == ManagedEnvironment.VSCODE_REMOTE_CONTAINERS:
            for k in os.environ.keys():
                if k.startswith('REMOTE_CONTAINERS_'):
                    self.environ[k] = os.environ[k]

    @staticmethod
    def check_container_call():
        container_v = os.environ.get('KAS_CONTAINER_SCRIPT_VERSION')
        if not container_v:
            # not a kas-container call (or from a too old script)
            return
        if container_v != __version__:
            logging.warning(f'kas-container ({container_v}) and '
                            f'kas ({__version__}) versions do not match')

    @staticmethod
    def _get_managed_env():
        """
            Detects if kas is running in well-known environment (e.g. a
            CI system). Returns the identifier of the CI system or None.
        """
        if os.environ.get('GITHUB_ACTIONS', False) == 'true':
            return ManagedEnvironment.GITHUB_ACTIONS
        if os.environ.get('GITLAB_CI', False) == 'true':
            return ManagedEnvironment.GITLAB_CI
        if os.environ.get('REMOTE_CONTAINERS', False) == 'true':
            return ManagedEnvironment.VSCODE_REMOTE_CONTAINERS
        return None

    @property
    def build_dir(self):
        """
            The path to the build directory
        """
        return self.__kas_build_dir

    @property
    def kas_work_dir(self):
        """
            The path to the kas work directory
        """
        return self.__kas_work_dir

    @property
    def kas_repo_ref_dir(self):
        """
            The reference directory for the repo
        """
        return self.__kas_repo_ref_dir

    @property
    def force_checkout(self):
        return getattr(self.args, 'force_checkout', None)

    @property
    def update(self):
        return getattr(self.args, 'update', None)

    @property
    def managed_env(self):
        return self._get_managed_env()



# kas - setup tool for bitbake based projects
#
# Copyright (c) Siemens AG, 2017-2024
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

'''
    This module contains the :kasschemadesc:`<node>` role to extract the
    description of a node from the schema. The `<node>` hereby is a path
    to the node in the schema, separated by dots. For example, the
    header.version node can be accessed with
    :kasschemadesc:`header.properties.version`. Definitions inside the
    `$defs` section are accessed by their full path, e.g.
    :kasschemadesc:`$defs.path.to.node`.
'''
__license__ = 'MIT'
__copyright__ = 'Copyright (c) Siemens AG, 2017-2024'

from sphinx.application import Sphinx
from sphinx.util.docutils import SphinxRole
from docutils.parsers.rst.states import Struct
from sphinx.util.nodes import nodes
from typing import Dict, Union
import re
from kas.configschema import CONFIGSCHEMA, __schema_definition__


class KasSchemaDescRole(SphinxRole):

    required_arguments = 1

    def __init__(self):
        super().__init__()
        self.key_regex = re.compile(r'([a-zA-Z0-9_]+)(?:\[(\d+)\])?')

    def run(self) -> tuple[list[nodes.Node], list[nodes.system_message]]:
        messages = []
        if self.text.startswith('$defs.'):
            node = CONFIGSCHEMA['$defs']
            path = self.text.split('.')[1:]
        else:
            node = CONFIGSCHEMA['properties']
            path = self.text.split('.')
        self.env.note_dependency(__schema_definition__)
        try:
            for part in path:
                match = self.key_regex.match(part).groups()
                if match[1]:
                    node = node[match[0]][int(match[1])]
                elif match[0]:
                    node = node[match[0]]
                else:
                    raise KeyError
        except KeyError:
            messages.append(self.inliner.document.reporter.error(
                f'Invalid path: {self.text}',
                line=self.lineno,
            ))
            return [], messages
        try:
            desc = node['description']
        except KeyError:
            messages.append(self.inliner.document.reporter.error(
                f'Description missing for path: {self.text}',
                line=self.lineno,
            ))
            return [], messages
        default = node.get('default', None)
        allowed_values = node.get('enum', None)
        minval = node.get('minimum', None)
        maxval = node.get('maximum', None)

        memo = Struct(document=self.inliner.document,
                      reporter=self.inliner.reporter,
                      language=self.inliner.language)
        parent = nodes.paragraph()
        processed, msgs = self.inliner.parse(desc, self.lineno, memo, parent)
        parent += processed
        messages += msgs
        if default is not None:
            def_pg = nodes.paragraph()
            def_pg += nodes.strong(text='Default: ')
            def_pg += nodes.literal(text=str(default))
            parent += def_pg
        if allowed_values is not None:
            av_pg = nodes.paragraph()
            av_pg += nodes.strong(text='Supported values: ')
            for i in range(len(allowed_values)):
                if i != 0:
                    av_pg += nodes.Text(', ')
                av_pg += nodes.literal(text=str(allowed_values[i]))
            parent += av_pg
        if minval is not None and maxval is not None:
            range_pg = nodes.paragraph()
            range_pg += nodes.strong(text='Range: ')
            range_pg += nodes.literal(text=f'[{minval}, {maxval}]')
            parent += range_pg

        return parent, messages


def setup(app: Sphinx) -> Dict[str, Union[bool, str]]:
    app.add_role('kasschemadesc', KasSchemaDescRole())

    return {
        'parallel_read_safe': True,
        'parallel_write_safe': True,
        'version': '1.0',
    }
